# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Geometry functions."""

from typing import Tuple

from brax import math
from brax.base import Contact
import jax
from jax import numpy as jp


def closest_segment_point(
    a: jp.ndarray, b: jp.ndarray, pt: jp.ndarray
) -> jp.ndarray:
  """Returns the closest point on the a-b line segment to a point pt."""
  ab = b - a
  t = jp.dot(pt - a, ab) / (jp.dot(ab, ab) + 1e-6)
  return a + jp.clip(t, 0.0, 1.0) * ab


def closest_segment_point_and_dist(
    a: jp.ndarray, b: jp.ndarray, pt: jp.ndarray
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Returns closest point on the line segment and the distance squared."""
  closest = closest_segment_point(a, b, pt)
  dist = (pt - closest).dot(pt - closest)
  return closest, dist


def closest_line_point(
    a: jp.ndarray, b: jp.ndarray, pt: jp.ndarray
) -> jp.ndarray:
  """Returns the closest point on the a-b line to a point pt."""
  ab = b - a
  t = jp.dot(pt - a, ab) / (jp.dot(ab, ab) + 1e-6)
  return a + t * ab


def closest_segment_to_segment_points(
    a0: jp.ndarray, a1: jp.ndarray, b0: jp.ndarray, b1: jp.ndarray
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Returns closest points on two line segments."""
  # Gets the closest segment points by first finding the closest points
  # between two lines. Points are then clipped to be on the line segments
  # and edge cases with clipping are handled.
  dir_a, len_a = math.normalize(a1 - a0)
  dir_b, len_b = math.normalize(b1 - b0)

  # Segment mid-points.
  half_len_a = len_a * 0.5
  half_len_b = len_b * 0.5
  a_mid = a0 + dir_a * half_len_a
  b_mid = b0 + dir_b * half_len_b

  # Translation between two segment mid-points.
  trans = a_mid - b_mid

  # Parametrize points on each line as follows:
  #  point_on_a = a_mid + t_a * dir_a
  #  point_on_b = b_mid + t_b * dir_b
  # and analytically minimize the distance between the two points.
  dira_dot_dirb = dir_a.dot(dir_b)
  dira_dot_trans = dir_a.dot(trans)
  dirb_dot_trans = dir_b.dot(trans)
  denom = 1 - dira_dot_dirb * dira_dot_dirb

  orig_t_a = (-dira_dot_trans + dira_dot_dirb * dirb_dot_trans) / (denom + 1e-6)
  orig_t_b = dirb_dot_trans + orig_t_a * dira_dot_dirb
  t_a = jp.clip(orig_t_a, -half_len_a, half_len_a)
  t_b = jp.clip(orig_t_b, -half_len_b, half_len_b)

  best_a = a_mid + dir_a * t_a
  best_b = b_mid + dir_b * t_b

  # Resolve edge cases where both closest points are clipped to the segment
  # endpoints by recalculating the closest segment points for the current
  # clipped points, and then picking the pair of points with smallest
  # distance. An example of this edge case is when lines intersect but line
  # segments don't.
  new_a, d1 = closest_segment_point_and_dist(a0, a1, best_b)
  new_b, d2 = closest_segment_point_and_dist(b0, b1, best_a)
  best_a = jp.where(d1 < d2, new_a, best_a)
  best_b = jp.where(d1 < d2, best_b, new_b)

  return best_a, best_b


def closest_segment_point_plane(
    a: jp.ndarray, b: jp.ndarray, p0: jp.ndarray, plane_normal: jp.ndarray
) -> jp.ndarray:
  """Gets the closest point between a line segment and a plane."""
  # If a line segment is parametrized as S(t) = a + t * (b - a), we can
  # plug it into the plane equation dot(n, S(t)) - d = 0, then solve for t to
  # get the line-plane intersection. We then clip t to be in [0, 1] to be on
  # the line segment.
  n = plane_normal
  d = jp.sum(p0 * n)  # shortest distance from origin to plane
  denom = jp.sum(n * (b - a))
  t = (d - jp.sum(n * a)) / (denom + 1e-6 * (denom == 0.0))
  t = jp.clip(t, 0, 1)
  segment_point = a + t * (b - a)

  return segment_point


def closest_triangle_point(
    p0: jp.ndarray, p1: jp.ndarray, p2: jp.ndarray, pt: jp.ndarray
) -> jp.ndarray:
  """Gets the closest point on a triangle to another point in space."""
  # Parametrize the triangle s.t. a point inside the triangle is
  # Q = p0 + u * e0 + v * e1, when 0 <= u <= 1, 0 <= v <= 1, and
  # 0 <= u + v <= 1. Let e0 = (p1 - p0) and e1 = (p2 - p0).
  # We analytically minimize the distance between the point pt and Q.
  e0 = p1 - p0
  e1 = p2 - p0
  a = e0.dot(e0)
  b = e0.dot(e1)
  c = e1.dot(e1)
  d = pt - p0
  # The determinant is 0 only if the angle between e1 and e0 is 0
  # (i.e. the triangle has overlapping lines).
  det = a * c - b * b
  u = (c * e0.dot(d) - b * e1.dot(d)) / det
  v = (-b * e0.dot(d) + a * e1.dot(d)) / det
  inside = (0 <= u) & (u <= 1) & (0 <= v) & (v <= 1) & (u + v <= 1)
  closest_p = p0 + u * e0 + v * e1
  d0 = (closest_p - pt).dot(closest_p - pt)

  # If the closest point is outside the triangle, it must be on an edge, so we
  # check each triangle edge for a closest point to the point pt.
  closest_p1, d1 = closest_segment_point_and_dist(p0, p1, pt)
  closest_p = jp.where((d0 < d1) & inside, closest_p, closest_p1)
  min_d = jp.where((d0 < d1) & inside, d0, d1)

  closest_p2, d2 = closest_segment_point_and_dist(p1, p2, pt)
  closest_p = jp.where(d2 < min_d, closest_p2, closest_p)
  min_d = jp.minimum(min_d, d2)

  closest_p3, d3 = closest_segment_point_and_dist(p2, p0, pt)
  closest_p = jp.where(d3 < min_d, closest_p3, closest_p)

  return closest_p


def closest_segment_triangle_points(
    a: jp.ndarray,
    b: jp.ndarray,
    p0: jp.ndarray,
    p1: jp.ndarray,
    p2: jp.ndarray,
    triangle_normal: jp.ndarray,
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Gets the closest points between a line segment and triangle."""
  # The closest triangle point is either on the edge or within the triangle.
  # First check triangle edges for the closest point.
  seg_pt1, tri_pt1 = closest_segment_to_segment_points(a, b, p0, p1)
  d1 = (seg_pt1 - tri_pt1).dot(seg_pt1 - tri_pt1)
  seg_pt2, tri_pt2 = closest_segment_to_segment_points(a, b, p1, p2)
  d2 = (seg_pt2 - tri_pt2).dot(seg_pt2 - tri_pt2)
  seg_pt3, tri_pt3 = closest_segment_to_segment_points(a, b, p0, p2)
  d3 = (seg_pt3 - tri_pt3).dot(seg_pt3 - tri_pt3)

  # Next, handle the case where the closest triangle point is inside the
  # triangle. Either the line segment intersects the triangle or a segment
  # endpoint is closest to a point inside the triangle.
  # If the line overlaps the triangle and is parallel to the triangle plane,
  # the chosen triangle point is arbitrary.
  seg_pt4 = closest_segment_point_plane(a, b, p0, triangle_normal)
  tri_pt4 = closest_triangle_point(p0, p1, p2, seg_pt4)
  d4 = (seg_pt4 - tri_pt4).dot(seg_pt4 - tri_pt4)

  # Get the point with minimum distance from the line segment point to the
  # triangle point.
  distance = jp.array([[d1, d2, d3, d4]])
  min_dist = jp.amin(distance)
  mask = (distance == min_dist).T
  seg_pt = jp.array([seg_pt1, seg_pt2, seg_pt3, seg_pt4]) * mask
  tri_pt = jp.array([tri_pt1, tri_pt2, tri_pt3, tri_pt4]) * mask
  seg_pt = jp.sum(seg_pt, axis=0) / jp.sum(mask)
  tri_pt = jp.sum(tri_pt, axis=0) / jp.sum(mask)

  return seg_pt, tri_pt


def project_pt_onto_plane(
    pt: jp.ndarray, plane_pt: jp.ndarray, plane_normal: jp.ndarray
) -> jp.ndarray:
  """Projects a point onto a plane along the plane normal."""
  dist = (pt - plane_pt).dot(plane_normal)
  return pt - dist * plane_normal


def _project_poly_onto_plane(
    poly: jp.ndarray, plane_pt: jp.ndarray, plane_normal: jp.ndarray
) -> jp.ndarray:
  """Projects a polygon onto a plane using the plane normal."""
  return jax.vmap(project_pt_onto_plane, in_axes=[0, None, None])(
      poly, plane_pt, math.normalize(plane_normal)[0]
  )


def _project_poly_onto_poly_plane(
    poly1: jp.ndarray, norm1: jp.ndarray, poly2: jp.ndarray, norm2: jp.ndarray
) -> jp.ndarray:
  """Projects poly1 onto the poly2 plane along poly1's normal."""
  d = poly2[0].dot(norm2)
  denom = norm1.dot(norm2)
  t = (d - poly1.dot(norm2)) / (denom + 1e-6 * (denom == 0.0))
  new_poly = poly1 + t.reshape(-1, 1) * norm1
  return new_poly


def point_in_front_of_plane(
    plane_pt: jp.ndarray, plane_normal: jp.ndarray, pt: jp.ndarray
) -> bool:
  """Checks if a point is strictly in front of a plane."""
  return (pt - plane_pt).dot(plane_normal) > 1e-6  # pytype: disable=bad-return-type  # jax-ndarray


def clip_edge_to_planes(
    edge_p0: jp.ndarray,
    edge_p1: jp.ndarray,
    plane_pts: jp.ndarray,
    plane_normals: jp.ndarray,
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Clips an edge against side planes.

  We return two clipped points, and a mask to include the new edge or not.

  Args:
    edge_p0: the first point on the edge
    edge_p1: the second point on the edge
    plane_pts: side plane points
    plane_normals: side plane normals

  Returns:
    new_ps: new edge points that are clipped against side planes
    mask: a boolean mask, True if an edge point is a valid clipped point and
    False otherwise
  """
  p0, p1 = edge_p0, edge_p1
  p0_in_front = jax.vmap(jp.dot)(p0 - plane_pts, plane_normals) > 1e-6
  p1_in_front = jax.vmap(jp.dot)(p1 - plane_pts, plane_normals) > 1e-6

  # Get candidate clipped points along line segment (p0, p1) by clipping against
  # all clipping planes.
  candidate_clipped_ps = jax.vmap(
      closest_segment_point_plane, in_axes=[None, None, 0, 0]
  )(p0, p1, plane_pts, plane_normals)

  def clip_edge_point(p0, p1, p0_in_front, clipped_ps):
    @jax.vmap
    def choose_edge_point(in_front, clipped_p):
      return jp.where(in_front, clipped_p, p0)

    # Pick the clipped point if p0 is in front of the clipping plane. Otherwise
    # keep p0 as the edge point.
    new_edge_ps = choose_edge_point(p0_in_front, clipped_ps)

    # Pick the clipped point that is most along the edge direction.
    # This degenerates to picking the original point p0 if p0 is *not* in front
    # of any clipping planes.
    dists = jp.dot(new_edge_ps - p0, p1 - p0)
    new_edge_p = new_edge_ps[jp.argmax(dists)]
    return new_edge_p

  # Clip each edge point.
  new_p0 = clip_edge_point(p0, p1, p0_in_front, candidate_clipped_ps)
  new_p1 = clip_edge_point(p1, p0, p1_in_front, candidate_clipped_ps)
  clipped_pts = jp.array([new_p0, new_p1])

  # Keep the original points if both points are in front of any of the clipping
  # planes, rather than creating a new clipped edge. If the entire subject edge
  # is in front of any clipping plane, we need to grab an edge from the clipping
  # polygon instead.
  both_in_front = p0_in_front & p1_in_front
  mask = ~jp.any(both_in_front)
  new_ps = jp.where(mask, clipped_pts, jp.array([p0, p1]))
  # Mask out crossing clipped edge points.
  mask = jp.where((p0 - p1).dot(new_ps[0] - new_ps[1]) < 0, False, mask)
  return new_ps, jp.array([mask, mask])


def clip(
    clipping_poly: jp.ndarray,
    subject_poly: jp.ndarray,
    clipping_normal: jp.ndarray,
    subject_normal: jp.ndarray,
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Clips a subject polygon against a clipping polygon.

  A parallelized clipping algorithm for convex polygons. The result is a set of
  vertices on the clipped subject polygon in the subject polygon plane.

  Args:
    clipping_poly: the polygon that we use to clip the subject polygon against
    subject_poly: the polygon that gets clipped
    clipping_normal: normal of the clipping polygon
    subject_normal: normal of the subject polygon

  Returns:
    clipped_pts: points on the clipped polygon
    mask: True if a point is in the clipping polygon, False otherwise
  """
  # Get clipping edge points, edge planes, and edge normals.
  clipping_p0 = jp.roll(clipping_poly, 1, axis=0)
  clipping_plane_pts = clipping_p0
  clipping_p1 = clipping_poly
  clipping_plane_normals = jax.vmap(jp.cross, in_axes=[0, None])(
      clipping_p1 - clipping_p0,
      clipping_normal,
  )

  # Get subject edge points, edge planes, and edge normals.
  subject_edge_p0 = jp.roll(subject_poly, 1, axis=0)
  subject_plane_pts = subject_edge_p0
  subject_edge_p1 = subject_poly
  subject_plane_normals = jax.vmap(jp.cross, in_axes=[0, None])(
      subject_edge_p1 - subject_edge_p0,
      subject_normal,
  )

  # Clip all edges of the subject poly against clipping side planes.
  clipped_edges0, masks0 = jax.vmap(
      clip_edge_to_planes, in_axes=[0, 0, None, None]
  )(
      subject_edge_p0,
      subject_edge_p1,
      clipping_plane_pts,
      clipping_plane_normals,
  )

  # Project the clipping poly onto the subject plane.
  clipping_p0_s = _project_poly_onto_poly_plane(
      clipping_p0, clipping_normal, subject_poly, subject_normal
  )
  # TODO consider doing a roll here instead of projection.
  clipping_p1_s = _project_poly_onto_poly_plane(
      clipping_p1, clipping_normal, subject_poly, subject_normal
  )

  # Clip all edges of the clipping poly against subject planes.
  clipped_edges1, masks1 = jax.vmap(
      clip_edge_to_planes, in_axes=[0, 0, None, None]
  )(clipping_p0_s, clipping_p1_s, subject_plane_pts, subject_plane_normals)

  # Merge the points and reshape.
  clipped_edges = jp.concatenate([clipped_edges0, clipped_edges1])
  masks = jp.concatenate([masks0, masks1])
  clipped_points = clipped_edges.reshape((-1, 3))
  mask = masks.reshape(-1)

  return clipped_points, mask


def manifold_points(
    poly: jp.ndarray, poly_mask: jp.ndarray, poly_norm: jp.ndarray
) -> jp.ndarray:
  """Chooses four points on the polygon with approximately maximal area."""
  dist_mask = jp.where(poly_mask, 0.0, -1e6)
  a_idx = jp.argmax(dist_mask)
  a = poly[a_idx]
  # choose point b furthest from a
  b_idx = (((a - poly) ** 2).sum(axis=1) + dist_mask).argmax()
  b = poly[b_idx]
  # choose point c furthest along the axis orthogonal to (a-b)
  ab = jp.cross(poly_norm, a - b)
  ap = a - poly
  c_idx = (jp.abs(ap.dot(ab)) + dist_mask).argmax()
  c = poly[c_idx]
  # choose point d furthest from the other two triangle edges
  ac = jp.cross(poly_norm, a - c)
  bc = jp.cross(poly_norm, b - c)
  bp = b - poly
  dist_bp = jp.abs(bp.dot(bc)) + dist_mask
  dist_ap = jp.abs(ap.dot(ac)) + dist_mask
  d_idx = jp.concatenate([dist_bp, dist_ap]).argmax() % poly.shape[0]
  return jp.array([a_idx, b_idx, c_idx, d_idx])


def _create_contact_manifold(
    clipping_poly: jp.ndarray,
    subject_poly: jp.ndarray,
    clipping_norm: jp.ndarray,
    subject_norm: jp.ndarray,
    sep_axis: jp.ndarray,
) -> Contact:
  """Creates a contact manifold between two convex polygons.

  The polygon faces are expected to have a counter clockwise winding order so
  that clipping plane normals point away from the polygon center.

  Args:
    clipping_poly: The reference polygon to clip the contact against.
    subject_poly: The subject polygon to clip contacts onto.
    clipping_norm: The clipping polygon normal.
    subject_norm: The subject polygon normal.
    sep_axis: The separating axis

  Returns:
    contact: Contact object.
  """
  # Clip the subject (incident) face onto the clipping (reference) face.
  # The incident points are clipped points on the subject polygon.
  poly_incident, mask = clip(
      clipping_poly, subject_poly, clipping_norm, subject_norm
  )
  # The reference points are clipped points on the clipping polygon.
  poly_ref = _project_poly_onto_plane(
      poly_incident, clipping_poly[0], clipping_norm
  )
  behind_clipping_plane = point_in_front_of_plane(
      clipping_poly[0], -clipping_norm, poly_incident
  )
  mask = mask & behind_clipping_plane

  # Choose four contact points.
  best = manifold_points(poly_ref, mask, clipping_norm)
  contact_pts = jp.take(poly_ref, best, axis=0)
  mask_pts = jp.take(mask, best, axis=0)
  penetration_dir = jp.take(poly_incident, best, axis=0) - contact_pts
  penetration = penetration_dir.dot(-clipping_norm)
  penetration = jp.where(mask_pts, penetration, -jp.ones_like(penetration))

  contact = Contact(
      pos=contact_pts,
      normal=jp.stack([sep_axis] * 4, 0),
      penetration=penetration,
      friction=jp.array([]),
      elasticity=jp.array([]),
      solver_params=jp.array([]),
      link_idx=jp.array([]),
  )

  return contact


def sat_hull_hull(
    faces_a: jp.ndarray,
    faces_b: jp.ndarray,
    vertices_a: jp.ndarray,
    vertices_b: jp.ndarray,
    normals_a: jp.ndarray,
    normals_b: jp.ndarray,
    unique_edges_a: jp.ndarray,
    unique_edges_b: jp.ndarray,
) -> Contact:
  """Runs the Separating Axis Test for a pair of hulls.

  Given two convex hulls, the Separating Axis Test finds a separating axis
  between all edge pairs and face pairs. Edge pairs create a single contact
  point and face pairs create a contact manifold (up to four contact points).
  We return both the edge and face contacts. Valid contacts can be checked with
  penetration > 0. Resulting edge contacts should be preferred over face
  contacts.

  Args:
    faces_a: An ndarray of hull A's polygon faces.
    faces_b: An ndarray of hull B's polygon faces.
    vertices_a: Vertices for hull A.
    vertices_b: Vertices for hull B.
    normals_a: Normal vectors for hull A's polygon faces.
    normals_b: Normal vectors for hull B's polygon faces.
    unique_edges_a: Unique edges for hull A.
    unique_edges_b: Unique edges for hull B.

  Returns:
    contact: A contact.
  """
  # get the separating axes
  edge_dir_a = unique_edges_a[:, 0] - unique_edges_a[:, 1]
  edge_dir_b = unique_edges_b[:, 0] - unique_edges_b[:, 1]
  edge_dir_a_r = jp.tile(edge_dir_a, reps=(unique_edges_b.shape[0], 1))
  edge_dir_b_r = jp.repeat(edge_dir_b, repeats=unique_edges_a.shape[0], axis=0)
  edge_edge_axes = jax.vmap(jp.cross)(edge_dir_a_r, edge_dir_b_r)
  edge_edge_axes = jax.vmap(lambda x: math.normalize(x, axis=0)[0])(
      edge_edge_axes
  )

  axes = jp.concatenate([normals_a, normals_b, edge_edge_axes])

  # for each separating axis, get the support
  @jax.vmap
  def get_support(axis):
    support_a = jax.vmap(jp.dot, in_axes=[None, 0])(axis, vertices_a)
    support_b = jax.vmap(jp.dot, in_axes=[None, 0])(axis, vertices_b)
    dist1 = support_a.max() - support_b.min()
    dist2 = support_b.max() - support_a.min()
    sign = jp.where(dist1 > dist2, -1, 1)
    dist = jp.minimum(dist1, dist2)
    dist = jp.where(~jp.all(axis == 0.0), dist, 1e6)  # degenerate axis
    return dist, sign

  support, sign = get_support(axes)

  # choose the best separating axis
  best_idx = jp.argmin(support)
  best_sign = sign[best_idx]
  best_axis = axes[best_idx]
  is_edge_contact = best_idx >= (normals_a.shape[0] + normals_b.shape[0])

  # get the (reference) face most aligned with the separating axis
  dist_a = jax.vmap(jp.dot, in_axes=[None, 0])(best_axis, normals_a)
  dist_b = jax.vmap(jp.dot, in_axes=[None, 0])(best_axis, normals_b)
  a_max = dist_a.argmax()
  b_max = dist_b.argmax()
  a_min = dist_a.argmin()
  b_min = dist_b.argmin()

  ref_face = jp.where(best_sign > 0, faces_a[a_max], faces_b[b_max])
  ref_face_norm = jp.where(best_sign > 0, normals_a[a_max], normals_b[b_max])
  incident_face = jp.where(best_sign > 0, faces_b[b_min], faces_a[a_min])
  incident_face_norm = jp.where(
      best_sign > 0, normals_b[b_min], normals_a[a_min]
  )

  contact = _create_contact_manifold(
      ref_face,
      incident_face,
      ref_face_norm,
      incident_face_norm,
      -best_sign * best_axis,
  )

  # For edge contacts, we use the clipped face point, mainly for performance
  # reasons. For small penetration, the clipped face point is roughly the edge
  # contact point.
  # TODO revisit edge contact pos (for deep penetration) with same perf
  idx = contact.penetration.argmax()
  contact = contact.replace(
      penetration=jp.where(
          is_edge_contact,
          jp.array([contact.penetration[idx], -1, -1, -1]),
          contact.penetration,
      ),
      pos=jp.where(
          is_edge_contact, jp.tile(contact.pos[idx], (4, 1)), contact.pos
      ),
  )

  return contact



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for geometry."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import math
from brax.geometry import math as geom_math
import jax
import jax.numpy as jp
import numpy as np
from scipy import spatial


def _get_rand_point(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(1, 3)
  return verts[0, :]


def _get_rand_line_segment(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(2, 3)
  return verts[0, :], verts[1, :]


def _get_rand_triangle_vertices(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(3, 3)
  return verts[0, :], verts[1, :], verts[2, :]


def _get_rand_dir(seed=None):
  if seed is not None:
    np.random.seed(seed)
  r = np.random.randn(1)
  theta = np.random.random(1) * 2 * np.pi
  a = (np.random.random(1) - 0.5) * 2.0
  phi = np.arccos(a)
  x = r * np.sin(phi) * np.cos(theta)
  y = r * np.sin(phi) * np.sin(theta)
  z = r * np.cos(phi)
  return jp.array([x, y, z]).squeeze()


def _get_rand_convex_polygon(seed=None):
  if seed is not None:
    np.random.seed(seed)
  points = np.random.random((15, 2))
  hull = spatial.ConvexHull(points)
  points = points[hull.vertices]
  points = np.array([[p[0], p[1], 0] for p in points])
  return points


def _minimize(fn, sample_fn, lb, ub, tol, max_iter=20, seed=42):
  """Minimize a function, roughly, using the cross-entropy method."""
  # We can alternatively use scipy.optimize with non-linear constraints here.
  assert lb.shape == ub.shape, 'bounds need to have the same shape'
  np.random.seed(seed)

  i, n = 0, 1_000
  mu = (ub + lb) * 0.5
  sigma = (ub - lb) * 0.5
  size = lb.shape[0]
  val, prev_val = fn(mu), None

  while prev_val is None or np.abs(val - prev_val) > tol:
    params = sample_fn(mu, sigma, n, size, lb, ub)
    vals = np.array([fn(p) for p in params])
    if val < vals.min():  # early exit
      return mu
    idx = vals.argsort()
    best_idx = idx[: int(n * 0.05)]
    mu = params[best_idx].mean(axis=0)
    sigma = params[best_idx].std(axis=0) + 1e-10

    prev_val = val
    val = fn(mu)

    i += 1
    if i == max_iter:
      break

  return mu


def _closest_segment_to_segment_points(a0, a1, b0, b1):
  dir_a = a1 - a0
  len_a = np.sqrt(dir_a.dot(dir_a))
  half_len_a = len_a / 2
  dir_a = dir_a / len_a

  dir_b = b1 - b0
  len_b = np.sqrt(dir_b.dot(dir_b))
  half_len_b = len_b / 2
  dir_b = dir_b / len_b

  a_mid = a0 + dir_a * half_len_a
  b_mid = b0 + dir_b * half_len_b

  # Parametrize both line segments.
  def fn(t):
    best_a = a_mid + dir_a * t[0]
    best_b = b_mid + dir_b * t[1]
    return (best_a - best_b).dot(best_a - best_b)

  def sample_fn(mu, sigma, n, size, lb, ub):
    params = np.random.normal(mu, sigma, size=(n, size))
    params = np.clip(params, lb, ub)
    return params

  lb = np.array([-half_len_a, -half_len_b])
  ub = np.array([half_len_a, half_len_b])
  ta, tb = _minimize(fn, sample_fn, lb, ub, tol=1e-4)
  best_a = a_mid + dir_a * ta
  best_b = b_mid + dir_b * tb
  return best_a, best_b


def _closest_triangle_point(p0, p1, p2, pt):
  # Parametrize the triangle using barycentric coordinates where a point in
  # the triangle Q = w * V_0 + y * V_1 + v * V_2
  # and u + v + w = 1, or w = 1 - (u + v), (u + v) <= 1.
  def fn(t):
    u, v = t[0], t[1]
    w = 1 - u - v
    best_q = u * p0 + v * p1 + w * p2
    return (pt - best_q).dot(pt - best_q)

  def sample_fn(mu, sigma, n, size, lb, ub):
    params = np.random.normal(mu, sigma, size=(n, size))
    params = np.clip(params, lb, ub)
    params[:, 1] = np.clip(params[:, 1], 0.0, 1.0 - params[:, 0])
    return params

  lb = np.array([0, 0])
  ub = np.array([1, 1])
  u, v = _minimize(fn, sample_fn, lb, ub, tol=1e-10)
  best_q = u * p0 + v * p1 + (1 - u - v) * p2
  return best_q


def _closest_segment_triangle_points(a, b, p0, p1, p2):
  # Parametrize line segment.
  dir_ls = b - a
  len_ls = jp.sqrt(dir_ls.dot(dir_ls))
  half_len_ls = len_ls / 2
  dir_ls = dir_ls / len_ls
  ls_mid = a + dir_ls * half_len_ls

  # Parametrize the triangle using barycentric coordinates.
  def fn(t):
    best_ls = ls_mid + dir_ls * t[0]
    u, v = t[1], t[2]
    w = 1 - u - v
    best_q = u * p0 + v * p1 + w * p2
    return (best_ls - best_q).dot(best_ls - best_q)

  def sample_fn(mu, sigma, n, size, lb, ub):
    params = np.random.normal(mu, sigma, size=(n, size))
    params = np.clip(params, lb, ub)
    params[:, 2] = np.clip(params[:, 2], 0.0, 1.0 - params[:, 1])
    return params

  lb = np.array([-half_len_ls, 0, 0])
  ub = np.array([half_len_ls, 1, 1])
  t_ls, u, v = _minimize(fn, sample_fn, lb, ub, tol=1e-10)
  best_ls = ls_mid + dir_ls * t_ls
  best_q = u * p0 + v * p1 + (1 - u - v) * p2
  return best_ls, best_q


class ClosestSegmentSegmentPointsTest(absltest.TestCase):
  """Tests for closest segment-to-segment points."""

  def test_closest_segments_points(self):
    a0 = jp.array([0.73432405, 0.12372768, 0.20272314])
    a1 = jp.array([1.10600128, 0.88555209, 0.65209485])
    b0 = jp.array([0.85599262, 0.61736299, 0.9843583])
    b1 = jp.array([1.84270939, 0.92891793, 1.36343326])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [1.09063, 0.85404, 0.63351], 5)
    self.assertSequenceAlmostEqual(best_b, [0.99596, 0.66156, 1.03813], 5)

  def test_intersecting_segments(self):
    """Tests segments that intersect."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    b0, b1 = jp.array([-1.0, 0.0, 0.0]), jp.array([1.0, 0.0, 0.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, 0.0], 5)
    self.assertSequenceAlmostEqual(best_b, [0.0, 0.0, 0.0], 5)

  def test_intersecting_lines(self):
    """Tests that intersecting lines (not segments) get clipped."""
    a0, a1 = jp.array([0.2, 0.2, 0.0]), jp.array([1.0, 1.0, 0.0])
    b0, b1 = jp.array([0.2, 0.4, 0.0]), jp.array([1.0, 2.0, 0.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.3, 0.3, 0.0], 2)
    self.assertSequenceAlmostEqual(best_b, [0.2, 0.4, 0.0], 2)

  def test_parallel_segments(self):
    """Tests that parallel segments have closest points at the midpoint."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    b0, b1 = jp.array([1.0, 0.0, -1.0]), jp.array([1.0, 0.0, 1.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, 0.0], 5)
    self.assertSequenceAlmostEqual(best_b, [1.0, 0.0, 0.0], 5)

  def test_parallel_offset_segments(self):
    """Tests that offset parallel segments are close at segment endpoints."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    b0, b1 = jp.array([1.0, 0.0, 1.0]), jp.array([1.0, 0.0, 3.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, 1.0], 5)
    self.assertSequenceAlmostEqual(best_b, [1.0, 0.0, 1.0], 5)

  def test_zero_length_segments(self):
    """Test that zero length segments don't return NaNs."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, -1.0])
    b0, b1 = jp.array([1.0, 0.0, 0.1]), jp.array([1.0, 0.0, 0.1])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, -1.0], 5)
    self.assertSequenceAlmostEqual(best_b, [1.0, 0.0, 0.1], 5)

  def test_overlapping_segments(self):
    """Tests that perfectly overlapping segments intersect at the midpoints."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    b0, b1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, 0.0], 5)
    self.assertSequenceAlmostEqual(best_b, [0.0, 0.0, 0.0], 5)


class GeometryScipyTest(parameterized.TestCase):
  """Tests for geometry functions against scipy equivalents."""

  params = list(zip(np.repeat(np.arange(10), 10), np.tile(np.arange(10), 10)))

  @parameterized.parameters(*params)
  def test_closest_segment_to_segment_points(self, i, j):
    a0, a1 = _get_rand_line_segment(i)
    b0, b1 = _get_rand_line_segment(j)
    expected = _closest_segment_to_segment_points(a0, a1, b0, b1)
    ans = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    expected_dist = (expected[0] - expected[1]).dot(expected[0] - expected[1])
    test_dist = (ans[0] - ans[1]).dot(ans[0] - ans[1])
    self.assertAlmostEqual(expected_dist, test_dist, 4)

  @parameterized.parameters(*params)
  def test_closest_triangle_point(self, i, j):
    pt = _get_rand_point(i)
    p0, p1, p2 = _get_rand_triangle_vertices(j)
    expected = _closest_triangle_point(p0, p1, p2, pt)
    ans = geom_math.closest_triangle_point(p0, p1, p2, pt)
    self.assertSequenceAlmostEqual(expected, ans, 4)

  @parameterized.parameters(*params)
  def test_closest_segment_triangle_points(self, i, j):
    a, b = _get_rand_line_segment(i)
    p0, p1, p2 = _get_rand_triangle_vertices(j)
    expected = _closest_segment_triangle_points(a, b, p0, p1, p2)
    triangle_normal, _ = math.normalize(jp.cross(p0 - p1, p2 - p1))
    ans = geom_math.closest_segment_triangle_points(
        a, b, p0, p1, p2, triangle_normal
    )
    expected_dist = (expected[0] - expected[1]).dot(expected[0] - expected[1])
    test_dist = (ans[0] - ans[1]).dot(ans[0] - ans[1])
    self.assertAlmostEqual(expected_dist, test_dist, 1)
    # Guarantee we are lower than an upper bound.
    self.assertLessEqual(test_dist, expected_dist + 1e-5)


def _check_eq_pts(pts1, pts2, atol=1e-6):
  # For every point in pts1, make sure we have a point in pts2
  # that is close within `atol`, and vice versa.
  if not pts1.size and not pts2.size:
    return True
  elif not pts1.size:
    return False
  elif not pts2.size:
    return False
  if pts1.shape[-1] != 3 or pts2.shape[-1] != 3:
    raise AssertionError('Points should be three dimensional.')
  eq = True
  for p1 in pts1:
    eq = eq and jp.any(jp.sum(jp.abs(p1 - pts2) < atol, axis=-1) == 3)
  for p2 in pts2:
    eq = eq and jp.any(jp.sum(jp.abs(p2 - pts1) < atol, axis=-1) == 3)
  return eq


def _clip(clipping_polygon, subject_polygon):
  """Returns the clipped subject polygon, using Sutherland-Hodgman Clipping."""
  polygon1 = clipping_polygon
  polygon1_normal = jp.cross(
      polygon1[-1] - polygon1[0], polygon1[1] - polygon1[0]
  )
  clipping_edges = [
      (polygon1[i - 1], polygon1[i]) for i in range(len(polygon1))
  ]
  # Clipping plane normals point away from the clipping poly center (polygon
  # points assumed to have a clockwise winding order).
  clipping_planes = [
      (e0, jp.cross(polygon1_normal, e1 - e0)) for e0, e1 in clipping_edges
  ]

  output_polygon = subject_polygon
  for clipping_plane in clipping_planes:
    input_ = output_polygon
    output_polygon = []

    starting_pt = input_[-1]
    clipping_plane_pt, clipping_plane_normal = clipping_plane

    for endpt in input_:
      intersection_pt = geom_math.closest_segment_point_plane(
          starting_pt, endpt, clipping_plane_pt, clipping_plane_normal
      )
      starting_pt_front = geom_math.point_in_front_of_plane(
          clipping_plane_pt, clipping_plane_normal, starting_pt
      )
      endpt_front = geom_math.point_in_front_of_plane(
          clipping_plane_pt, clipping_plane_normal, endpt
      )

      if not endpt_front and not starting_pt_front:
        output_polygon.append(endpt)
      elif not endpt_front and starting_pt_front:
        output_polygon.append(intersection_pt)
        output_polygon.append(endpt)
      elif not starting_pt_front:
        output_polygon.append(intersection_pt)

      starting_pt = endpt

    if not output_polygon:
      # All clipping points outside the subject polygon.
      return jp.array(output_polygon)

  return jp.array(output_polygon)


class ClippingTest(parameterized.TestCase):
  """Tests that the clipping algorithm matches a baseline implementation."""

  clip_vectorized = jax.jit(geom_math.clip)

  @parameterized.parameters(range(100))
  def test_clipped_triangles(self, i):
    subject_poly = jp.array(_get_rand_triangle_vertices(i))
    clipping_poly = jp.array(_get_rand_triangle_vertices(i + 1))

    poly_out = _clip(clipping_poly, subject_poly)

    clipping_normal = jp.cross(
        clipping_poly[1] - clipping_poly[0],
        clipping_poly[-1] - clipping_poly[0],
    )
    subject_normal = jp.cross(
        subject_poly[1] - subject_poly[0], subject_poly[-1] - subject_poly[0]
    )
    poly_out_jax, mask = ClippingTest.clip_vectorized(
        clipping_poly, subject_poly, clipping_normal, subject_normal
    )

    self.assertTrue(
        _check_eq_pts(poly_out, poly_out_jax[mask], atol=1e-4),
        f'Clipped triangles {i} did not match.',
    )

  @parameterized.parameters(range(100))
  def test_clipped_hulls(self, i):
    # The hulls are all in the x-y plane, unlike in `test_clipped_triangles`.
    subject_poly = jp.array(_get_rand_convex_polygon(i))
    clipping_poly = jp.array(_get_rand_convex_polygon(i + 1))

    poly_out = _clip(clipping_poly, subject_poly)

    clipping_normal = jp.cross(
        clipping_poly[1] - clipping_poly[0],
        clipping_poly[-1] - clipping_poly[0],
    )
    subject_normal = jp.cross(
        subject_poly[1] - subject_poly[0],
        subject_poly[-1] - subject_poly[0],
    )
    poly_out_jax, mask = ClippingTest.clip_vectorized(
        clipping_poly, subject_poly, clipping_normal, subject_normal
    )

    self.assertTrue(
        _check_eq_pts(poly_out, poly_out_jax[mask], atol=1e-2),
        f'Clipped hulls {i} did not match.',
    )


class ManifoldPointsTest(parameterized.TestCase):
  """Tests manifold point selection."""

  def test_manifold_points(self):
    poly = jp.array([
        [0.99999994, 0.14842263, 0.39985055],
        [0.8585786, 0.00145163, 0.39985055],
        [1.0, -0.14551926, 0.39985055],
        [1.1414213, 0.00145174, 0.39985055],
    ])
    poly_mask = jp.array([False, True, True, True])
    poly_norm = jp.array([0.0, 0.0, 1.0])
    idx = geom_math.manifold_points(poly, poly_mask, poly_norm)
    self.assertSequenceEqual(idx.tolist(), [1, 3, 2, 1])


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Useful functions for creating and processing meshes."""

import itertools
import logging
from typing import Dict, Tuple, Union

from brax.base import Box, Convex, Mesh
from jax import numpy as jp
import numpy as np
from scipy import spatial
import trimesh


_BOX_CORNERS = list(itertools.product((-1, 1), (-1, 1), (-1, 1)))

# pyformat: disable
# The faces of a triangulated box, i.e. the indices in _BOX_CORNERS of the
# vertices of the 12 triangles (two triangles for each side of the box).
_TRIANGULATED_BOX_FACES = [
    0, 4, 1, 4, 5, 1,  # left
    0, 2, 4, 2, 6, 4,  # bottom
    6, 5, 4, 6, 7, 5,  # front
    2, 3, 6, 3, 7, 6,  # right
    1, 5, 3, 5, 7, 3,  # top
    0, 1, 2, 1, 3, 2,  # back
]
# Rectangular box faces using a counter-clockwise winding order convention.
_BOX_FACES = [
    0, 4, 5, 1,  # left
    0, 2, 6, 4,  # bottom
    6, 7, 5, 4,  # front
    2, 3, 7, 6,  # right
    1, 5, 7, 3,  # top
    0, 1, 3, 2,  # back
]
# pyformat: enable

_MAX_HULL_FACE_VERTICES = 20
_CONVEX_CACHE: Dict[Tuple[int, int], Convex] = {}


def get_face_norm(vert: jp.ndarray, face: jp.ndarray) -> jp.ndarray:
  """Calculates face normals given vertices and face indexes."""
  assert len(vert.shape) == 2 and len(face.shape) == 2, (
      f'vert and face should have dim of 2, got {len(vert.shape)} and '
      f'{len(face.shape)}'
  )
  face_vert = jp.take(vert, face, axis=0)
  # use CCW winding order convention
  edge0 = face_vert[:, 1, :] - face_vert[:, 0, :]
  edge1 = face_vert[:, -1, :] - face_vert[:, 0, :]
  face_norm = jp.cross(edge0, edge1)
  face_norm = face_norm / jp.linalg.norm(face_norm, axis=1).reshape((-1, 1))
  return face_norm


def get_unique_edges(vert: np.ndarray, face: np.ndarray) -> np.ndarray:
  """Returns unique edges.

  Args:
    vert: (n_vert, 3) vertices
    face: (n_face, n_vert) face index array

  Returns:
    edges: 2-tuples of vertex indexes for each edge
  """
  r_face = np.roll(face, 1, axis=1)
  edges = np.concatenate(np.array([face, r_face]).T)

  # do a first pass to remove duplicates
  edges.sort(axis=1)
  edges = np.unique(edges, axis=0)
  edges = edges[edges[:, 0] != edges[:, 1]]  # get rid of edges from padded face

  # get normalized edge directions
  edge_vert = vert.take(edges, axis=0)
  edge_dir = edge_vert[:, 0] - edge_vert[:, 1]
  norms = np.sqrt(np.sum(edge_dir**2, axis=1))
  edge_dir = edge_dir / norms.reshape((-1, 1))

  # get the first unique edge for all pairwise comparisons
  diff1 = edge_dir[:, None, :] - edge_dir[None, :, :]
  diff2 = edge_dir[:, None, :] + edge_dir[None, :, :]
  matches = (np.linalg.norm(diff1, axis=-1) < 1e-6) | (
      np.linalg.norm(diff2, axis=-1) < 1e-6
  )
  matches = np.tril(matches).sum(axis=-1)
  unique_edge_idx = np.where(matches == 1)[0]

  return edges[unique_edge_idx]


def _box(b: Box, triangulated=True) -> Tuple[np.ndarray, np.ndarray]:
  """Creates face and vert from a box geometry."""
  assert b.halfsize.shape == (
      3,
  ), f'Box halfsize should have shape (3,), got {b.halfsize.shape}'
  box_corners = jp.array(_BOX_CORNERS)
  vert = box_corners * b.halfsize.reshape(-1, 3)
  box_faces = _TRIANGULATED_BOX_FACES if triangulated else _BOX_FACES
  face_dim = 3 if triangulated else 4
  face = jp.array([box_faces]).reshape(-1, face_dim)
  return vert, face


def box_tri(b: Box) -> Mesh:
  """Creates a triangulated mesh from a box geometry."""
  vert, face = _box(b, triangulated=True)
  return Mesh(  # pytype: disable=wrong-arg-types  # jax-ndarray
      vert=vert,
      face=face,
      link_idx=b.link_idx,
      transform=b.transform,
      friction=b.friction,
      elasticity=b.elasticity,
      solver_params=b.solver_params,
      rgba=b.rgba,
  )


def _box_hull(b: Box) -> Convex:
  """Creates a mesh for a box with rectangular faces."""
  vert, face = _box(b, triangulated=False)
  return Convex(  # pytype: disable=wrong-arg-types  # jax-ndarray
      vert=vert,
      face=face,
      link_idx=b.link_idx,
      transform=b.transform,
      friction=b.friction,
      elasticity=b.elasticity,
      solver_params=b.solver_params,
      unique_edge=get_unique_edges(vert, face),
      rgba=b.rgba,
  )


def convex_hull_2d(points: np.ndarray, normal: np.ndarray) -> np.ndarray:
  """Calculates the hull face for a set of points on a plane."""
  # project points onto the closest axis plane
  best_axis = np.abs(np.eye(3).dot(normal)).argmax()
  axis = np.eye(3)[best_axis]
  d = points.dot(axis).reshape((-1, 1))
  axis_points = points - d * axis
  axis_points = axis_points[:, list({0, 1, 2} - {best_axis})]

  # get the polygon hull face, and make the points ccw wrt the face normal
  # TODO: consider sorting unique edges by their angle to get the hull
  c = spatial.ConvexHull(axis_points)
  order_ = np.where(axis.dot(normal) > 0, 1, -1)
  order_ *= np.where(best_axis == 1, -1, 1)
  hull_point_idx = c.vertices[::order_]
  assert (axis_points - c.points).sum() == 0

  return hull_point_idx


def _merge_coplanar(tm: trimesh.Trimesh) -> np.ndarray:
  """Merges coplanar facets."""
  if not tm.facets:
    return tm.faces.copy()  # no facets, return faces
  if not tm.faces.shape[0]:
    raise ValueError('Mesh has no faces.')

  # Get faces.
  face_idx = set(range(tm.faces.shape[0])) - set(np.concatenate(tm.facets))
  face_idx = np.array(list(face_idx))
  faces = tm.faces[face_idx] if face_idx.shape[0] > 0 else np.array([])

  # Get facets.
  facets = []
  for i, facet in enumerate(tm.facets):
    point_idx = np.unique(tm.faces[facet])
    points = tm.vertices[point_idx]
    normal = tm.facets_normal[i]

    # convert triangulated facet to a polygon
    hull_point_idx = convex_hull_2d(points, normal)
    face = point_idx[hull_point_idx]

    # resize faces that exceed max polygon vertices
    every = face.shape[0] // _MAX_HULL_FACE_VERTICES + 1
    face = face[::every]
    facets.append(face)

  # Pad facets so that they can be stacked.
  max_len = max(f.shape[0] for f in facets) if facets else faces.shape[1]
  assert max_len <= _MAX_HULL_FACE_VERTICES
  for i, f in enumerate(facets):
    if f.shape[0] < max_len:
      f = np.pad(f, (0, max_len - f.shape[0]), 'edge')
    facets[i] = f

  if not faces.shape[0]:
    assert facets
    return np.array(facets)  # no faces, return facets

  # Merge faces and facets.
  faces = np.pad(faces, ((0, 0), (0, max_len - faces.shape[1])), 'edge')
  return np.concatenate([faces, facets])


def _convex_hull(m: Mesh) -> Convex:
  """Creates a convex hull from a mesh."""
  tm = trimesh.Trimesh(vertices=m.vert, faces=m.face)
  tm_convex = trimesh.convex.convex_hull(tm)
  vert = tm_convex.vertices.copy()
  face = _merge_coplanar(tm_convex)
  return Convex(  # pytype: disable=wrong-arg-types  # jax-ndarray
      vert=vert,
      face=face,
      link_idx=m.link_idx,
      transform=m.transform,
      friction=m.friction,
      elasticity=m.elasticity,
      solver_params=m.solver_params,
      unique_edge=get_unique_edges(vert, face),
      rgba=m.rgba,
  )


def convex_hull(obj: Union[Box, Mesh]) -> Convex:
  """Creates a convex hull from a box or mesh."""
  if isinstance(obj, Box):
    return _box_hull(obj)
  key = (hash(obj.vert.data.tobytes()), hash(obj.face.data.tobytes()))  # pytype: disable=attribute-error  # jax-ndarray
  if key not in _CONVEX_CACHE:
    logging.info('Converting mesh %s into convex hull.', key)
    _CONVEX_CACHE[key] = _convex_hull(obj)
  convex = _CONVEX_CACHE[key]
  convex = convex.replace(
      link_idx=obj.link_idx,
      transform=obj.transform,
      friction=obj.friction,
      elasticity=obj.elasticity,
      solver_params=obj.solver_params,
      rgba=obj.rgba,
  )
  return convex



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for mesh.py."""

from absl.testing import absltest
from brax.base import Box, Convex, Mesh, Transform
from brax.geometry import mesh
import numpy as np


class BoxTest(absltest.TestCase):

  def test_box(self):
    """Tests a triangulated box."""
    b = Box(
        halfsize=np.repeat(0.5, 3),
        link_idx=None,
        transform=None,
        friction=0.42,
        elasticity=1,
        solver_params=None,
    )
    m = mesh.box_tri(b)
    self.assertIsInstance(m, Mesh)
    self.assertSequenceEqual(m.vert.shape, (8, 3))  # eight box corners
    self.assertEqual(np.unique(np.abs(m.vert)), 0.5)
    self.assertSequenceEqual(m.face.shape, (12, 3))  # two triangles per face
    self.assertEqual(m.friction, 0.42)

    expected_face_norm = [
        [0, -1.0, 0],  # left
        [0, -1.0, 0],
        [0, 0, -1.0],  # bottom
        [0, 0, -1.0],
        [+1.0, 0, 0],  # front
        [+1.0, 0, 0],
        [0, +1.0, 0],  # right
        [0, +1.0, 0],
        [0, 0, +1.0],  # top
        [0, 0, +1.0],
        [-1.0, 0, 0],  # back
        [-1.0, 0, 0],
    ]
    face_norm = mesh.get_face_norm(m.vert, m.face)
    np.testing.assert_array_almost_equal(face_norm, expected_face_norm)

  def test_box_hull(self):
    """Tests a polygon box."""
    b = Box(
        halfsize=np.repeat(0.5, 3).reshape(3),
        link_idx=None,
        transform=None,
        friction=0.42,
        elasticity=1,
        solver_params=None,
    )
    h = mesh.convex_hull(b)
    self.assertIsInstance(h, Convex)
    self.assertSequenceEqual(h.vert.shape, (8, 3))
    self.assertEqual(np.unique(np.abs(h.vert)), 0.5)
    np.testing.assert_array_equal(h.unique_edge, [[0, 1], [0, 2], [0, 4]])
    self.assertSequenceEqual(h.face.shape, (6, 4))  # one rectangle per face
    self.assertEqual(h.friction, 0.42)

    expected_face_norm = [
        [0, -1.0, 0],  # left
        [0, 0, -1.0],  # bottom
        [+1.0, 0, 0],  # front
        [0, +1.0, 0],  # right
        [0, 0, +1.0],  # top
        [-1.0, 0, 0],  # back
    ]
    face_norm = mesh.get_face_norm(h.vert, h.face)
    np.testing.assert_array_almost_equal(face_norm, expected_face_norm)


class ConvexTest(absltest.TestCase):

  def test_pyramid(self):
    """Tests a pyramid convex hull."""
    vert = np.array([
        [-0.025, 0.05, 0.05],
        [-0.025, -0.05, -0.05],
        [-0.025, -0.05, 0.05],
        [-0.025, 0.05, -0.05],
        [0.075, 0.0, 0.0],
    ])
    face = np.array(
        [[0, 1, 2], [0, 3, 1], [0, 4, 3], [0, 2, 4], [2, 1, 4], [1, 3, 4]]
    )
    pyramid = Mesh(
        link_idx=0,
        transform=Transform.zero((1,)),
        vert=vert,
        face=face,
        friction=1,
        elasticity=0,
        solver_params=None,
    )
    h = mesh.convex_hull(pyramid)

    self.assertIsInstance(h, Convex)

    # check verts
    vidx = [0, 3, 2, 1, 4]  # verts get mixed up by trimesh
    np.testing.assert_array_equal(h.vert, vert[vidx])

    # check faces
    map_ = {v: k for k, v in enumerate(vidx)}
    h_face = np.vectorize(map_.get)(h.face)
    np.testing.assert_array_equal(
        h_face,
        np.array([
            [2, 4, 0, 0],
            [0, 4, 3, 3],
            [4, 2, 1, 1],
            [1, 3, 4, 4],
            [3, 1, 2, 0],
        ]),
    )

    # check edges
    np.testing.assert_array_equal(
        np.vectorize(map_.get)(h.unique_edge),
        np.array([[0, 3], [0, 2], [0, 4], [3, 4], [2, 4], [1, 4]]),
    )
    self.assertEqual(h.friction, 1)


class ConvexHull2DTest(absltest.TestCase):

  def test_convex_hull_2d_axis1(self):
    """Tests for the correct winding order for a polgyon with +y normal."""
    pts = np.array([
        [-0.04634297, -0.06652775, 0.05853534],
        [-0.01877651, -0.08309858, -0.05236476],
        [0.02362804, -0.08010745, 0.05499557],
        [0.04066505, -0.09034877, -0.01354446],
        [-0.07255043, -0.06837638, -0.00781699],
    ])
    normal = np.array([-0.18467607, -0.97768016, 0.10018111])
    idx = mesh.convex_hull_2d(pts, normal)
    expected = np.cross(pts[idx][1] - pts[idx][0], pts[idx][2] - pts[idx][0])
    expected /= np.linalg.norm(expected)
    np.testing.assert_array_almost_equal(normal, expected)

  def test_convex_hull_2d_axis2(self):
    """Tests for the correct winding order for a polgyon with +z normal."""
    pts = np.array([
        [0.08607829, -0.03881998, -0.03291714],
        [-0.01877651, -0.08309858, -0.05236476],
        [0.05470364, 0.00027677, -0.08371042],
        [-0.01010019, -0.02708892, -0.0957297],
        [0.04066505, -0.09034877, -0.01354446],
    ])
    normal = np.array([0.3839915, -0.60171936, -0.70034587])
    idx = mesh.convex_hull_2d(pts, normal)
    expected = np.cross(pts[idx][1] - pts[idx][0], pts[idx][2] - pts[idx][0])
    expected /= np.linalg.norm(expected)
    np.testing.assert_array_almost_equal(normal, expected)


class UniqueEdgesTest(absltest.TestCase):

  def test_tetrahedron_edges(self):
    """Tests unique edges for a tetrahedron."""
    vert = np.array(
        [[-0.1, 0.0, -0.1], [0.0, 0.1, 0.1], [0.1, 0.0, -0.1], [0.0, -0.1, 0.1]]
    )
    face = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 1], [2, 1, 3]])
    idx = mesh.get_unique_edges(vert, face)
    np.testing.assert_array_equal(
        idx, np.array([[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]])
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for image."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import test_utils
from brax.io import image
from brax.spring import pipeline
import jax
import jax.numpy as jp


class ImageTest(parameterized.TestCase):

  @parameterized.parameters([('ant.xml',), ('convex_convex.xml',)])
  def test_render_array(self, xml):
    sys = test_utils.load_fixture(xml)
    state = jax.jit(pipeline.init)(sys, sys.init_q, jp.zeros(sys.qd_size()))
    im = image.render_array(sys, state, 32, 32)
    self.assertEqual(im.shape, (32, 32, 3))

if __name__ == '__main__':
  absltest.main()



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example: a Component."""
AUTHORS = ('Shixiang Shane Gu',)
CONTACTS = ('shanegu@google.com',)
AFFILIATIONS = ('google.com',)
DESCRIPTIONS = ('ant with four legs',)

ENVS = {}

COMPONENTS = dict(
    ant=dict(
        module='ant',
        tracks=('race', 'race_ma', 'sumo', 'onigokko_predator',
                'onigokko_prey'),
    ),)



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exports a system config and state as an image."""

import io
from typing import List, Optional, Tuple

import brax
from brax import base
from brax import math
import jax
from jax import numpy as jp
import numpy as onp
from PIL import Image
from pytinyrenderer import TinyRenderCamera as Camera
from pytinyrenderer import TinyRenderLight as Light
from pytinyrenderer import TinySceneRenderer as Renderer
import trimesh


class TextureRGB888:

  def __init__(self, pixels):
    self.pixels = pixels
    self.width = int(onp.sqrt(len(pixels) / 3))
    self.height = int(onp.sqrt(len(pixels) / 3))


class Grid(TextureRGB888):

  def __init__(self, grid_size, color):
    grid = onp.zeros((grid_size, grid_size, 3), dtype=onp.int32)
    grid[:, :] = onp.array(color)
    grid[0] = onp.zeros((grid_size, 3), dtype=onp.int32)
    grid[:, 0] = onp.zeros((grid_size, 3), dtype=onp.int32)
    super().__init__(list(grid.ravel()))


_BASIC = TextureRGB888([133, 118, 102])
_TARGET = TextureRGB888([255, 34, 34])
_GROUND = Grid(100, [200, 200, 200])


def _scene(sys: brax.System, state: brax.State) -> Tuple[Renderer, List[int]]:
  """Converts a brax System and state to a pytinyrenderer scene and instances."""
  scene = Renderer()
  instances = []

  def take_i(obj, i):
    return jax.tree_map(lambda x: jp.take(x, i, axis=0), obj)

  link_names = [n or f'link {i}' for i, n in enumerate(sys.link_names)]
  link_names += ['world']
  link_geoms = {}
  for batch in sys.geoms:
    num_geoms = len(batch.friction)
    for i in range(num_geoms):
      link_idx = -1 if batch.link_idx is None else batch.link_idx[i]
      link_geoms.setdefault(link_names[link_idx], []).append(take_i(batch, i))

  for _, geom in link_geoms.items():
    for col in geom:
      tex = TextureRGB888((col.rgba[:3] * 255).astype('uint32'))
      if isinstance(col, base.Capsule):
        half_height = col.length / 2
        model = scene.create_capsule(col.radius, half_height, 2,
                                     tex.pixels, tex.width, tex.height)
      elif isinstance(col, base.Box):
        model = scene.create_cube(col.halfsize, tex.pixels, tex.width,
                                  tex.height, 16.)
      elif isinstance(col, base.Sphere):
        model = scene.create_capsule(
            col.radius, 0, 2, tex.pixels, tex.width, tex.height
        )
      elif isinstance(col, base.Plane):
        tex = _GROUND
        model = scene.create_cube([1000.0, 1000.0, 0.0001], tex.pixels,
                                  tex.width, tex.height, 8192)
      elif isinstance(col, base.Convex):
        # convex objects are not visual
        continue
      elif isinstance(col, base.Mesh):
        tm = trimesh.Trimesh(vertices=col.vert, faces=col.face)
        vert_norm = tm.vertex_normals
        model = scene.create_mesh(
            col.vert.reshape((-1)).tolist(),
            vert_norm.reshape((-1)).tolist(),
            [0] * col.vert.shape[0] * 2,
            col.face.reshape((-1)).tolist(),
            tex.pixels,
            tex.width,
            tex.height,
            1.0,
        )
      else:
        raise RuntimeError(f'unrecognized collider: {type(col)}')

      i = col.link_idx if col.link_idx is not None else -1
      x = state.x.concatenate(base.Transform.zero((1,)))
      instance = scene.create_object_instance(model)
      off = col.transform.pos
      pos = onp.array(x.pos[i]) + math.rotate(off, x.rot[i])
      rot = col.transform.rot
      rot = math.quat_mul(x.rot[i], rot)
      scene.set_object_position(instance, list(pos))
      scene.set_object_orientation(instance, [rot[1], rot[2], rot[3], rot[0]])
      instances.append(instance)

  return scene, instances


def _eye(sys: brax.System, state: brax.State) -> List[float]:
  """Determines the camera location for a Brax system."""
  xj = state.x.vmap().do(sys.link.joint)
  dist = onp.concatenate(xj.pos[None, ...] - xj.pos[:, None, ...])
  dist = onp.linalg.norm(dist, axis=1).max()
  off = [2 * dist, -2 * dist, dist]
  return list(state.x.pos[0, :] + onp.array(off))


def _up(unused_sys: brax.System) -> List[float]:
  """Determines the up orientation of the camera."""
  return [0, 0, 1]


def get_camera(
    sys: brax.System, state: brax.State, width: int, height: int, ssaa: int = 2
) -> Camera:
  """Gets camera object."""
  eye, up = _eye(sys, state), _up(sys)
  hfov = 58.0
  vfov = hfov * height / width
  target = [state.x.pos[0, 0], state.x.pos[0, 1], 0]
  camera = Camera(
      viewWidth=width * ssaa,
      viewHeight=height * ssaa,
      position=eye,
      target=target,
      up=up,
      hfov=hfov,
      vfov=vfov)
  return camera


def render_array(sys: brax.System,
                 state: brax.State,
                 width: int,
                 height: int,
                 light: Optional[Light] = None,
                 camera: Optional[Camera] = None,
                 ssaa: int = 2) -> onp.ndarray:
  """Renders an RGB array of a brax system and QP."""
  if (len(state.x.pos.shape), len(state.x.rot.shape)) != (2, 2):
    raise RuntimeError('unexpected shape in state')
  scene, instances = _scene(sys, state)
  target = state.x.pos[0, :]
  if light is None:
    direction = [0.57735, -0.57735, 0.57735]
    light = Light(
        direction=direction,
        ambient=0.8,
        diffuse=0.8,
        specular=0.6,
        shadowmap_center=target)
  if camera is None:
    eye, up = _eye(sys, state), _up(sys)
    hfov = 58.0
    vfov = hfov * height / width
    camera = Camera(
        viewWidth=width * ssaa,
        viewHeight=height * ssaa,
        position=eye,
        target=target,
        up=up,
        hfov=hfov,
        vfov=vfov)
  img = scene.get_camera_image(instances, light, camera).rgb
  arr = onp.reshape(
      onp.array(img, dtype=onp.uint8),
      (camera.view_height, camera.view_width, -1))
  if ssaa > 1:
    arr = onp.asarray(Image.fromarray(arr).resize((width, height)))
  return arr


def render(sys: brax.System,
           states: List[brax.State],
           width: int,
           height: int,
           light: Optional[Light] = None,
           cameras: Optional[List[Camera]] = None,
           ssaa: int = 2,
           fmt='png') -> bytes:
  """Returns an image of a brax system and QP."""
  if not states:
    raise RuntimeError('must have at least one qp')
  if cameras is None:
    cameras = [None] * len(states)

  frames = [
      Image.fromarray(
          render_array(sys, state, width, height, light, camera, ssaa))
      for state, camera in zip(states, cameras)
  ]
  f = io.BytesIO()
  if len(frames) == 1:
    frames[0].save(f, format=fmt)
  else:
    frames[0].save(
        f,
        format=fmt,
        append_images=frames[1:],
        save_all=True,
        duration=sys.dt * 1000,
        loop=0)
  return f.getvalue()



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Calculations for generating contacts."""

from typing import Iterator, Optional, Tuple

from brax import math
from brax.base import (
    Capsule,
    Cylinder,
    Contact,
    Convex,
    Geometry,
    Mesh,
    Plane,
    Sphere,
    System,
    Transform,
)
from brax.geometry import math as geom_math
from brax.geometry import mesh as geom_mesh
import jax
from jax import numpy as jp


def _sphere_plane(sphere: Sphere, plane: Plane) -> Contact:
  """Calculates one contact between a sphere and a plane."""
  n = math.rotate(jp.array([0.0, 0.0, 1.0]), plane.transform.rot)
  t = jp.dot(sphere.transform.pos - plane.transform.pos, n)
  penetration = sphere.radius - t
  # halfway between contact points on sphere and on plane
  pos = sphere.transform.pos - n * (sphere.radius - 0.5 * penetration)
  c = Contact(pos, n, penetration, *_combine(sphere, plane))
  # returns 1 contact, so add a batch dimension of size 1
  return jax.tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _sphere_sphere(s_a: Sphere, s_b: Sphere) -> Contact:
  """Calculates one contact between two spheres."""
  n, dist = math.normalize(s_a.transform.pos - s_b.transform.pos)
  penetration = s_a.radius + s_b.radius - dist
  s_a_pos = s_a.transform.pos - n * s_a.radius
  s_b_pos = s_b.transform.pos + n * s_b.radius
  pos = (s_a_pos + s_b_pos) * 0.5
  c = Contact(pos, n, penetration, *_combine(s_a, s_b))
  # returns 1 contact, so add a batch dimension of size 1
  return jax.tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _sphere_capsule(sphere: Sphere, capsule: Capsule) -> Contact:
  """Calculates one contact between a sphere and a capsule."""
  segment = jp.array([0.0, 0.0, capsule.length * 0.5])
  segment = math.rotate(segment, capsule.transform.rot)
  pt = geom_math.closest_segment_point(
      capsule.transform.pos - segment,
      capsule.transform.pos + segment,
      sphere.transform.pos,
  )
  n, dist = math.normalize(sphere.transform.pos - pt)
  penetration = sphere.radius + capsule.radius - dist

  sphere_pos = sphere.transform.pos - n * sphere.radius
  cap_pos = pt + n * capsule.radius
  pos = (sphere_pos + cap_pos) * 0.5

  c = Contact(pos, n, penetration, *_combine(sphere, capsule))
  # returns 1 contact, so add a batch dimension of size 1
  return jax.tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _sphere_circle(sphere: Sphere, circle: Cylinder) -> Contact:
  """Calculates one contact between a sphere and a circle."""
  n = math.rotate(jp.array([0.0, 0.0, 1.0]), circle.transform.rot)

  # orient the normal s.t. it points at the CoM of the sphere
  normal_dir = jp.sign(
      (sphere.transform.pos - circle.transform.pos).dot(n))
  n = n * normal_dir

  pos = sphere.transform.pos - n * sphere.radius
  plane_pt = circle.transform.pos
  penetration = jp.dot(plane_pt - pos, n)

  # check if the sphere radius is within the cylinder in the normal dir of the
  # circle
  plane_pt2 = plane_pt + n
  line_pt = geom_math.closest_line_point(
      plane_pt, plane_pt2, sphere.transform.pos
  )
  in_cylinder = (sphere.transform.pos - line_pt).dot(
      sphere.transform.pos - line_pt
  ) <= circle.radius**2

  # get closest point on circle edge
  perp_dir = jp.cross(n, sphere.transform.pos - plane_pt)
  perp_dir = math.rotate(perp_dir, math.quat_rot_axis(n, -jp.pi / 2.0))
  perp_dir, _ = math.normalize(perp_dir)
  edge_pt = plane_pt + perp_dir * circle.radius
  edge_contact = (sphere.transform.pos - edge_pt).dot(
      sphere.transform.pos - edge_pt
  ) <= sphere.radius**2
  edge_to_sphere = sphere.transform.pos - edge_pt
  edge_to_sphere = math.normalize(edge_to_sphere)[0]

  penetration = jp.where(in_cylinder, penetration, -jp.ones_like(penetration))
  penetration = jp.where(
      edge_contact,
      sphere.radius
      - jp.sqrt(
          (sphere.transform.pos - edge_pt).dot(sphere.transform.pos - edge_pt)
      ),
      penetration,
  )
  n = jp.where(edge_contact, edge_to_sphere, n)
  pos = jp.where(edge_contact, edge_pt, pos)
  c = Contact(pos, n, penetration, *_combine(sphere, circle))  # pytype: disable=wrong-arg-types  # jax-ndarray
  return jax.tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _sphere_convex(sphere: Sphere, convex: Convex) -> Contact:
  """Calculates contacts between a sphere and a convex object."""
  # Get convex transformed normals, faces, and vertices.
  normals = geom_mesh.get_face_norm(convex.vert, convex.face)
  faces = jp.take(convex.vert, convex.face, axis=0)

  @jax.vmap
  def transform_faces(face, normal):
    face = convex.transform.pos + jax.vmap(math.rotate, in_axes=[0, None])(
        face, convex.transform.rot
    )
    normal = math.rotate(normal, convex.transform.rot)
    return face, normal

  faces, normals = transform_faces(faces, normals)

  # Get support from face normals.
  @jax.vmap
  def get_support(faces, normal):
    sphere_pos = sphere.transform.pos - normal * sphere.radius
    return jp.dot(sphere_pos - faces[0], normal)

  support = get_support(faces, normals)

  # Pick the face with minimal penetration as long as it has support.
  support = jp.where(support >= 0, -1e12, support)
  best_idx = support.argmax()
  face = faces[best_idx]
  normal = normals[best_idx]

  # Get closest point between the polygon face and the sphere center point.
  # Project the sphere center point onto poly plane. If it's inside polygon
  # edge normals, then we're done.
  pt = geom_math.project_pt_onto_plane(sphere.transform.pos, face[0], normal)
  edge_p0 = jp.roll(face, 1, axis=0)
  edge_p1 = face
  edge_normals = jax.vmap(jp.cross, in_axes=[0, None])(
      edge_p1 - edge_p0,
      normal,
  )
  edge_dist = jax.vmap(
      lambda plane_pt, plane_norm: (pt - plane_pt).dot(plane_norm)
  )(edge_p0, edge_normals)
  inside = jp.all(edge_dist <= 0)  # lte to handle degenerate edges

  # If the point is outside edge normals, project onto the closest edge plane
  # that the point is in front of.
  degenerate_edge = jp.all(edge_normals == 0, axis=1)
  behind = edge_dist < 0.0
  edge_dist = jp.where(degenerate_edge | behind, 1e12, edge_dist)
  idx = edge_dist.argmin()
  edge_pt = geom_math.closest_segment_point(edge_p0[idx], edge_p1[idx], pt)

  pt = jp.where(inside, pt, edge_pt)

  # Get the normal, penetration, and contact position.
  n, d = math.normalize(sphere.transform.pos - pt)
  spt = sphere.transform.pos - n * sphere.radius
  penetration = sphere.radius - d
  pos = (pt + spt) * 0.5

  c = Contact(pos, n, penetration, *_combine(sphere, convex))
  return jax.tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _sphere_mesh(sphere: Sphere, mesh: Mesh) -> Contact:
  """Calculates contacts between a sphere and a mesh."""

  @jax.vmap
  def sphere_face(face):
    pt = mesh.transform.pos + jax.vmap(math.rotate, in_axes=[0, None])(
        face, mesh.transform.rot
    )
    p0, p1, p2 = pt[0, :], pt[1, :], pt[2, :]

    tri_p = geom_math.closest_triangle_point(p0, p1, p2, sphere.transform.pos)
    n = sphere.transform.pos - tri_p
    n, dist = math.normalize(n)
    penetration = sphere.radius - dist
    sph_p = sphere.transform.pos - n * sphere.radius
    pos = (tri_p + sph_p) * 0.5
    return Contact(pos, n, penetration, *_combine(sphere, mesh))

  return sphere_face(jp.take(mesh.vert, mesh.face, axis=0))


def _capsule_plane(capsule: Capsule, plane: Plane) -> Contact:
  """Calculates two contacts between a capsule and a plane."""
  segment = jp.array([0.0, 0.0, capsule.length * 0.5])
  segment = math.rotate(segment, capsule.transform.rot)

  results = []
  for off in [segment, -segment]:
    sphere = Sphere(
        radius=capsule.radius,
        link_idx=capsule.link_idx,
        transform=Transform.create(pos=capsule.transform.pos + off),
        friction=capsule.friction,
        elasticity=capsule.elasticity,
        solver_params=capsule.solver_params,
    )
    results.append(_sphere_plane(sphere, plane))

  return jax.tree_map(lambda *x: jp.concatenate(x), *results)


def _capsule_capsule(cap_a: Capsule, cap_b: Capsule) -> Contact:
  """Calculates one contact between two capsules."""
  seg_a = jp.array([0.0, 0.0, cap_a.length * 0.5])
  seg_a = math.rotate(seg_a, cap_a.transform.rot)
  seg_b = jp.array([0.0, 0.0, cap_b.length * 0.5])
  seg_b = math.rotate(seg_b, cap_b.transform.rot)
  pt_a, pt_b = geom_math.closest_segment_to_segment_points(
      cap_a.transform.pos - seg_a,
      cap_a.transform.pos + seg_a,
      cap_b.transform.pos - seg_b,
      cap_b.transform.pos + seg_b,
  )
  n, dist = math.normalize(pt_a - pt_b)
  penetration = cap_a.radius + cap_b.radius - dist

  cap_a_pos = pt_a - n * cap_a.radius
  cap_b_pos = pt_b + n * cap_b.radius
  pos = (cap_a_pos + cap_b_pos) * 0.5

  c = Contact(pos, n, penetration, *_combine(cap_a, cap_b))
  # returns 1 contact, so add a batch dimension of size 1
  return jax.tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _capsule_convex(capsule: Capsule, convex: Convex) -> Contact:
  """Calculates contacts between a capsule and a convex object."""
  # Get convex transformed normals, faces, and vertices.
  normals = geom_mesh.get_face_norm(convex.vert, convex.face)
  faces = jp.take(convex.vert, convex.face, axis=0)

  @jax.vmap
  def transform_faces(faces, normals):
    faces = convex.transform.pos + jax.vmap(math.rotate, in_axes=[0, None])(
        faces, convex.transform.rot
    )
    normals = math.rotate(normals, convex.transform.rot)
    return faces, normals

  faces, normals = transform_faces(faces, normals)

  seg = jp.array([0.0, 0.0, capsule.length * 0.5])
  seg = math.rotate(seg, capsule.transform.rot)
  cap_pts = jp.array([
      capsule.transform.pos - seg,
      capsule.transform.pos + seg,
  ])

  # Get support from face normals.
  @jax.vmap
  def get_support(face, normal):
    pts = cap_pts - normal * capsule.radius
    sup = jax.vmap(lambda x: jp.dot(x - face[0], normal))(pts)
    return sup.min()

  support = get_support(faces, normals)
  has_support = jp.all(support < 0)

  # Pick the face with minimal penetration as long as it has support.
  support = jp.where(support >= 0, -1e12, support)
  best_idx = support.argmax()
  face = faces[best_idx]
  normal = normals[best_idx]

  # Clip the edge against side planes and create two contact points against the
  # face.
  edge_p0 = jp.roll(face, 1, axis=0)
  edge_p1 = face
  edge_normals = jax.vmap(jp.cross, in_axes=[0, None])(
      edge_p1 - edge_p0,
      normal,
  )
  cap_pts_clipped, mask = geom_math.clip_edge_to_planes(
      cap_pts[0], cap_pts[1], edge_p0, edge_normals
  )
  cap_pts_clipped = cap_pts_clipped - normal * capsule.radius
  face_pts = jax.vmap(geom_math.project_pt_onto_plane, in_axes=[0, None, None])(
      cap_pts_clipped, face[0], normal
  )
  # Create variables for the face contact.
  pos = (cap_pts_clipped + face_pts) * 0.5
  norm = jp.stack([normal] * 2, 0)
  penetration = jp.where(
      mask & has_support, jp.dot(face_pts - cap_pts_clipped, normal), -1
  )

  # Get a potential edge contact.
  # TODO handle deep edge penetration more gracefully, since edge_axis
  # can point in the wrong direction for deep penetration.
  edge_closest, cap_closest = jax.vmap(
      geom_math.closest_segment_to_segment_points, in_axes=[0, 0, None, None]
  )(edge_p0, edge_p1, cap_pts[0], cap_pts[1])
  e_idx = ((edge_closest - cap_closest) ** 2).sum(axis=1).argmin()
  cap_closest_pt, edge_closest_pt = cap_closest[e_idx], edge_closest[e_idx]
  edge_axis = cap_closest_pt - edge_closest_pt
  edge_axis, edge_dist = math.normalize(edge_axis)
  edge_pos = (
      edge_closest_pt + (cap_closest_pt - edge_axis * capsule.radius)
  ) * 0.5
  edge_norm = edge_axis
  edge_penetration = capsule.radius - edge_dist
  has_edge_contact = edge_penetration > 0

  # Create the contact.
  pos = jp.where(has_edge_contact, pos.at[0].set(edge_pos), pos)
  norm = jp.where(has_edge_contact, norm.at[0].set(edge_norm), norm)
  penetration = jp.where(
      has_edge_contact, penetration.at[0].set(edge_penetration), penetration
  )
  tile_fn = lambda x: jp.tile(x, (2,) + tuple([1 for _ in x.shape]))
  params = jax.tree_map(tile_fn, _combine(capsule, convex))
  return Contact(pos, norm, penetration, *params)


def _capsule_mesh(capsule: Capsule, mesh: Mesh) -> Contact:
  """Calculates contacts between a capsule and a mesh."""

  @jax.vmap
  def capsule_face(face, face_norm):
    seg = jp.array([0.0, 0.0, capsule.length * 0.5])
    seg = math.rotate(seg, capsule.transform.rot)
    end_a, end_b = capsule.transform.pos - seg, capsule.transform.pos + seg

    tri_norm = math.rotate(face_norm, mesh.transform.rot)
    pt = mesh.transform.pos + jax.vmap(math.rotate, in_axes=[0, None])(
        face, mesh.transform.rot
    )
    p0, p1, p2 = pt[..., 0, :], pt[..., 1, :], pt[..., 2, :]

    seg_p, tri_p = geom_math.closest_segment_triangle_points(
        end_a, end_b, p0, p1, p2, tri_norm
    )
    n = seg_p - tri_p
    n, dist = math.normalize(n)
    penetration = capsule.radius - dist
    cap_p = seg_p - n * capsule.radius
    pos = (tri_p + cap_p) * 0.5
    return Contact(pos, n, penetration, *_combine(capsule, mesh))

  face_vert = jp.take(mesh.vert, mesh.face, axis=0)
  face_norm = geom_mesh.get_face_norm(mesh.vert, mesh.face)
  return capsule_face(face_vert, face_norm)


def _convex_plane(convex: Convex, plane: Plane) -> Contact:
  """Calculates contacts between a convex object and a plane."""

  @jax.vmap
  def transform_verts(vertices):
    return convex.transform.pos + math.rotate(vertices, convex.transform.rot)

  vertices = transform_verts(convex.vert)
  n = math.rotate(jp.array([0.0, 0.0, 1.0]), plane.transform.rot)
  support = jax.vmap(jp.dot, in_axes=[None, 0])(
      n, plane.transform.pos - vertices
  )
  idx = geom_math.manifold_points(vertices, support > 0, n)

  pos = vertices[idx]
  normal = jp.stack([n] * 4, axis=0)
  unique = jp.tril(idx == idx[:, None]).sum(axis=1) == 1
  penetration = jp.where(unique, support[idx], -1)
  tile_fn = lambda x: jp.tile(x, (4,) + tuple([1 for _ in x.shape]))
  params = jax.tree_map(tile_fn, _combine(convex, plane))
  return Contact(pos, normal, penetration, *params)


def _convex_convex(convex_a: Convex, convex_b: Convex) -> Contact:
  """Calculates contacts between two convex objects."""
  # pad face vertices so that we can broadcast between geom_i and geom_j
  sa, sb = convex_a.face.shape[-1], convex_b.face.shape[-1]
  if sa < sb:
    face = jp.pad(convex_a.face, ((0, sb - sa)), 'edge')
    convex_a = convex_a.replace(face=face)
  elif sb < sa:
    face = jp.pad(convex_b.face, ((0, sa - sb)), 'edge')
    convex_b = convex_b.replace(face=face)

  normals_a = geom_mesh.get_face_norm(convex_a.vert, convex_a.face)
  normals_b = geom_mesh.get_face_norm(convex_b.vert, convex_b.face)
  faces_a = jp.take(convex_a.vert, convex_a.face, axis=0)
  faces_b = jp.take(convex_b.vert, convex_b.face, axis=0)

  def transform_faces(convex, faces, normals):
    faces = convex.transform.pos + jax.vmap(math.rotate, in_axes=[0, None])(
        faces, convex.transform.rot
    )
    normals = math.rotate(normals, convex.transform.rot)
    return faces, normals

  v_transform_faces = jax.vmap(transform_faces, in_axes=[None, 0, 0])
  faces_a, normals_a = v_transform_faces(convex_a, faces_a, normals_a)
  faces_b, normals_b = v_transform_faces(convex_b, faces_b, normals_b)

  def transform_verts(convex, vertices):
    vertices = convex.transform.pos + math.rotate(
        vertices, convex.transform.rot
    )
    return vertices

  v_transform_verts = jax.vmap(transform_verts, in_axes=[None, 0])
  vertices_a = v_transform_verts(convex_a, convex_a.vert)
  vertices_b = v_transform_verts(convex_b, convex_b.vert)

  unique_edges_a = jp.take(vertices_a, convex_a.unique_edge, axis=0)
  unique_edges_b = jp.take(vertices_b, convex_b.unique_edge, axis=0)

  c = geom_math.sat_hull_hull(
      faces_a,
      faces_b,
      vertices_a,
      vertices_b,
      normals_a,
      normals_b,
      unique_edges_a,
      unique_edges_b,
  )
  tile_fn = lambda x: jp.tile(x, (4,) + tuple([1 for _ in x.shape]))
  params = jax.tree_map(tile_fn, _combine(convex_a, convex_b))

  return Contact(c.pos, c.normal, c.penetration, *params)


def _mesh_plane(mesh: Mesh, plane: Plane) -> Contact:
  """Calculates contacts between a mesh and a plane."""

  @jax.vmap
  def point_plane(vert):
    n = math.rotate(jp.array([0.0, 0.0, 1.0]), plane.transform.rot)
    pos = mesh.transform.pos + math.rotate(vert, mesh.transform.rot)
    penetration = jp.dot(plane.transform.pos - pos, n)
    return Contact(pos, n, penetration, *_combine(mesh, plane))

  return point_plane(mesh.vert)


def _combine(
    geom_a: Geometry, geom_b: Geometry
) -> Tuple[jp.ndarray, jp.ndarray, jp.ndarray, Tuple[jp.ndarray, jp.ndarray]]:
  # default is to take maximum, but can override
  friction = jp.maximum(geom_a.friction, geom_b.friction)
  elasticity = (geom_a.elasticity + geom_b.elasticity) * 0.5
  solver_params = (geom_a.solver_params + geom_b.solver_params) * 0.5
  link_idx = (
      jp.array(geom_a.link_idx if geom_a.link_idx is not None else -1),
      jp.array(geom_b.link_idx if geom_b.link_idx is not None else -1),
  )
  return friction, elasticity, solver_params, link_idx


_TYPE_FUN = {
    (Sphere, Plane): _sphere_plane,
    (Sphere, Sphere): _sphere_sphere,
    (Sphere, Capsule): _sphere_capsule,
    (Sphere, Cylinder): _sphere_circle,
    (Sphere, Convex): _sphere_convex,
    (Sphere, Mesh): _sphere_mesh,
    (Capsule, Plane): _capsule_plane,
    (Capsule, Capsule): _capsule_capsule,
    (Capsule, Convex): _capsule_convex,
    (Capsule, Mesh): _capsule_mesh,
    (Convex, Convex): _convex_convex,
    (Convex, Plane): _convex_plane,
    (Mesh, Plane): _mesh_plane,
}


def _geom_pairs(sys: System) -> Iterator[Tuple[Geometry, Geometry]]:
  for i in range(len(sys.geoms)):
    for j in range(i, len(sys.geoms)):
      mask_i, mask_j = sys.geom_masks[i], sys.geom_masks[j]
      if (mask_i & mask_j >> 32) | (mask_i >> 32 & mask_j) == 0:
        continue
      geom_i, geom_j = sys.geoms[i], sys.geoms[j]
      if i == j and geom_i.link_idx is None:
        continue
      if i == j and geom_j.transform.pos.shape[0] == 1:
        continue
      yield (geom_i, geom_j)


def contact(sys: System, x: Transform) -> Optional[Contact]:
  """Calculates contacts in the system.

  Args:
    sys: system defining the kinematic tree and other properties
    x: link transforms in world frame

  Returns:
    Contact pytree, one row for each element in sys.contacts

  Raises:
    RuntimeError: if sys.contacts has an invalid type pair
  """

  contacts = []
  for geom_i, geom_j in _geom_pairs(sys):
    key = (type(geom_i), type(geom_j))
    fun = _TYPE_FUN.get(key)
    if fun is None:
      geom_i, geom_j = geom_j, geom_i
      fun = _TYPE_FUN.get((key[1], key[0]))
      if fun is None:
        raise RuntimeError(f'unrecognized collider pair: {key}')
    tx_i = x.take(geom_i.link_idx).vmap().do(geom_i.transform)

    if geom_i is geom_j:
      geom_i = geom_i.replace(transform=tx_i)
      choose_i, choose_j = jp.triu_indices(geom_i.link_idx.shape[0], 1)
      geom_i, geom_j = geom_i.take(choose_i), geom_i.take(choose_j)
      c = jax.vmap(fun)(geom_i, geom_j)  # type: ignore
      c = jax.tree_map(jp.concatenate, c)
    else:
      geom_i = geom_i.replace(transform=tx_i)
      # OK for geom j to have no parent links, e.g. static terrain
      if geom_j.link_idx is not None:
        tx_j = x.take(geom_j.link_idx).vmap().do(geom_j.transform)
        geom_j = geom_j.replace(transform=tx_j)
      vvfun = jax.vmap(jax.vmap(fun, in_axes=(0, None)), in_axes=(None, 0))
      c = vvfun(geom_i, geom_j)  # type: ignore
      c = jax.tree_map(lambda x: jp.concatenate(jp.concatenate(x)), c)

    contacts.append(c)

  if not contacts:
    return None

  # ignore penetration of two geoms within the same link
  c = jax.tree_map(lambda *x: jp.concatenate(x), *contacts)
  penetration = jp.where(c.link_idx[0] != c.link_idx[1], c.penetration, -1)
  c = c.replace(penetration=penetration)

  return c



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for mjcf."""

from absl.testing import absltest
from brax import test_utils
from brax.base import Box, Convex, Plane, Sphere, Mesh
import numpy as np

assert_almost_equal = np.testing.assert_array_almost_equal


class MjcfTest(absltest.TestCase):

  def test_load_pendulum(self):
    sys = test_utils.load_fixture('triple_pendulum.xml')

    assert_almost_equal(sys.gravity, np.array([0, 0, -9.81]))

    # check links
    self.assertSequenceEqual(sys.link_names, ['body1', 'body2', 'body3'])
    assert_almost_equal(
        sys.link.transform.pos, np.array([[0, 0, 0], [0, 0.5, 0], [0, 0.5, 0]])
    )
    assert_almost_equal(
        sys.link.transform.rot,
        np.array([[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0]]),
    )
    assert_almost_equal(
        sys.link.inertia.i, np.tile(np.eye(3), (3, 1, 1)) * 0.009
    )
    assert_almost_equal(
        sys.link.inertia.transform.pos,
        np.array([[0, 0.5, 0], [0, 0.5, 0], [0, 0.5, 0]]),
    )
    assert_almost_equal(sys.link.inertia.mass, np.array([1, 1, 1]))

    # check static fields: link types, parents
    self.assertEqual(sys.link_types, '111')
    self.assertEqual(sys.link_parents, (-1, 0, 1))

  def test_load_ant(self):
    sys = test_utils.load_fixture('ant.xml')

    # check links
    self.assertSequenceEqual(
        sys.link_names,
        ['torso', 'aux_1', '', 'aux_2', '', 'aux_3', '', 'aux_4', ''],
    )

    # check static fields: link types, parents
    self.assertEqual(sys.link_types, 'f11111111')
    self.assertEqual(sys.link_parents, (-1, 0, 1, 0, 3, 0, 5, 0, 7))

    # check contacts
    plane, _, _, sphere = sys.geoms
    self.assertIsInstance(sphere, Sphere)
    self.assertIsInstance(plane, Plane)
    assert_almost_equal(
        sphere.transform.pos,
        np.array(([
            [0.4, 0.4, 0.0],
            [-0.4, 0.4, 0.0],
            [-0.4, -0.4, 0.0],
            [0.4, -0.4, 0.0],
        ])),
    )
    assert_almost_equal(plane.transform.pos, np.array([[0, 0, 0]]))
    assert_almost_equal(sphere.link_idx, np.array([2, 4, 6, 8]))
    self.assertIsNone(plane.link_idx)

  def test_load_humanoid(self):
    sys = test_utils.load_fixture('humanoid.xml')

    # check links
    self.assertSequenceEqual(
        sys.link_names,
        [
            'torso',
            'lwaist',
            'pelvis',
            'right_thigh',
            'right_shin',
            'left_thigh',
            'left_shin',
            'right_upper_arm',
            'right_lower_arm',
            'left_upper_arm',
            'left_lower_arm',
        ],
    )
    self.assertEqual(sys.link_types, 'f2131312121')

    # check contacts
    plane, _, _, sphere = sys.geoms
    self.assertIsInstance(sphere, Sphere)
    self.assertIsInstance(plane, Plane)
    assert_almost_equal(
        sphere.transform.pos, np.array(([[0, 0, -0.35], [0, 0, -0.35]]))
    )
    assert_almost_equal(plane.transform.pos, np.array([[0, 0, 0]]))
    assert_almost_equal(sphere.link_idx, np.array([4, 6]))
    self.assertIsNone(plane.link_idx)

  def test_load_mesh_and_box(self):
    sys = test_utils.load_fixture('convex_convex.xml')
    n_box, n_convex, n_mesh = 0, 0, 0
    for g in sys.geoms:
      if isinstance(g, Box):
        n_box += 1
      elif isinstance(g, Convex):
        n_convex += 1
      elif isinstance(g, Mesh):
        n_mesh += 1
    self.assertEqual(n_box, 1)
    self.assertEqual(n_convex, 4)
    self.assertEqual(n_mesh, 3)

  def test_load_urdf(self):
    sys = test_utils.load_fixture('laikago/laikago_toes_zup.urdf')
    n_meshes = sum(
        g.friction.shape[0] if isinstance(g, Mesh) else 0 for g in sys.geoms
    )
    n_convex = sum(
        g.friction.shape[0] if isinstance(g, Convex) else 0 for g in sys.geoms
    )
    self.assertEqual(n_meshes, 26)
    self.assertEqual(n_convex, 26)

  def test_custom(self):
    sys = test_utils.load_fixture('capsule.xml')
    self.assertSequenceEqual([g.elasticity for g in sys.geoms], [0.2, 0.1])

  def test_joint_ref_check(self):
    with self.assertRaisesRegex(NotImplementedError, '`ref` attribute'):
      test_utils.load_fixture('nonzero_joint_ref.xml')

  def test_rgba(self):
    sys = test_utils.load_fixture('colour_objects.xml')
    # non default colour in plane
    assert_almost_equal(
        sys.geoms[0].rgba.squeeze(), np.array([1, 0, 0.8, 1]), 6
    )
    # other geometris have default colour
    for g in sys.geoms[1:]:
      assert_almost_equal(g.rgba.squeeze(), np.array([0.8, 0.6, 0.4, 1.0]), 6)

  def test_world_body_transform(self):
    sys = test_utils.load_fixture('world_body_transform.xml')
    # world body is in the right position/orientation
    r = 0.70710677
    assert_almost_equal(sys.geoms[1].transform.pos, np.array([[1.0, 0.0, 0.0]]))
    assert_almost_equal(
        sys.geoms[1].transform.rot, np.array([[r, 0.0, r, 0.0]]), 5
    )
    # child body is transformed wrt world body
    assert_almost_equal(
        sys.init_q, np.array([1.245, 0.0, 0.0, 0.5, 0.5, 0.5, -0.5])
    )

  def test_load_flat_cylinder(self):
    sys = test_utils.load_fixture('flat_cylinder.xml')
    self.assertEqual(sys.geoms[1].radius, 0.25)
    self.assertEqual(sys.geoms[1].length, 0.002)

  def test_load_fat_cylinder(self):
    with self.assertRaisesRegex(
        NotImplementedError, 'Cylinders of half-length'
    ):
      test_utils.load_fixture('fat_cylinder.xml')

  def test_load_fluid_box(self):
    sys = test_utils.load_fixture('fluid_box.xml')
    assert_almost_equal(sys.density, 1.2)
    assert_almost_equal(sys.viscosity, 0.15)

  def test_load_fluid_ellipsoid(self):
    with self.assertRaisesRegex(
        NotImplementedError, 'Ellipsoid fluid model not implemented'
    ):
      test_utils.load_fixture('fluid_ellipsoid.xml')

  def test_load_wind(self):
    with self.assertRaisesRegex(
        NotImplementedError, 'option.wind is not implemented'
    ):
      test_utils.load_fixture('fluid_wind.xml')


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Function to load MuJoCo mjcf format to Brax system."""

import itertools
from typing import Dict, Optional, Tuple, Union
import warnings
from xml.etree import ElementTree

from brax import math
from brax.base import (
    Actuator,
    Box,
    Capsule,
    Cylinder,
    DoF,
    Inertia,
    Link,
    Mesh,
    Motion,
    Plane,
    Sphere,
    System,
    Transform,
)
from brax.geometry import mesh as geom_mesh
from etils import epath
import jax
from jax import numpy as jp
import mujoco
import numpy as np


def _transform_do(
    pos: np.ndarray, quat: np.ndarray, cpos: np.ndarray, cquat: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
  pos = pos + math.rotate_np(cpos, quat)
  rot = math.quat_mul_np(quat, cquat)
  return pos, rot


def _fuse_bodies(elem: ElementTree.Element):
  """Fuses together parent child bodies that have no joint."""

  for child in list(elem):  # we will modify elem children, so make a copy
    _fuse_bodies(child)
    # this only applies to bodies with no joints
    if child.tag != 'body':
      continue
    if child.find('joint') is not None or child.find('freejoint') is not None:
      continue
    cpos = child.attrib.get('pos', '0 0 0')
    cpos = np.fromstring(cpos, sep=' ')
    cquat = child.attrib.get('quat', '1 0 0 0')
    cquat = np.fromstring(cquat, sep=' ')
    for grandchild in child:
      # TODO: might need to offset more than just body, geom
      if grandchild.tag in ('body', 'geom') and (cpos != 0).any():
        gcpos = grandchild.attrib.get('pos', '0 0 0')
        gcquat = grandchild.attrib.get('quat', '1 0 0 0')
        gcpos = np.fromstring(gcpos, sep=' ')
        gcquat = np.fromstring(gcquat, sep=' ')
        gcpos, gcquat = _transform_do(cpos, cquat, gcpos, gcquat)
        gcpos = ' '.join('%f' % i for i in gcpos)
        gcquat = ' '.join('%f' % i for i in gcquat)
        grandchild.attrib['pos'] = gcpos
        grandchild.attrib['quat'] = gcquat
      elem.append(grandchild)
    elem.remove(child)


def _get_meshdir(elem: ElementTree.Element) -> Union[str, None]:
  """Gets the mesh directory specified by the mujoco compiler tag."""
  elems = list(elem.iter('compiler'))
  return elems[0].get('meshdir') if elems else None


def _find_assets(
    elem: ElementTree.Element,
    path: epath.Path,
    meshdir: Optional[str],
) -> Dict[str, bytes]:
  """Loads assets from an xml given a base path."""
  assets = {}
  path = path if path.is_dir() else path.parent
  fname = elem.attrib.get('file') or elem.attrib.get('filename')
  if fname and fname.endswith('.xml'):
    # an asset can be another xml!  if so, we must traverse it, too
    asset = (path / fname).read_text()
    asset_xml = ElementTree.fromstring(asset)
    _fuse_bodies(asset_xml)
    asset_meshdir = _get_meshdir(asset_xml)
    assets[fname] = ElementTree.tostring(asset_xml)
    assets.update(_find_assets(asset_xml, path, asset_meshdir))
  elif fname:
    # mesh, png, etc
    path = path / meshdir if meshdir else path
    assets[fname] = (path / fname).read_bytes()

  for child in list(elem):
    assets.update(_find_assets(child, path, meshdir))

  return assets


def _get_mesh(mj: mujoco.MjModel, i: int) -> Tuple[np.ndarray, np.ndarray]:
  """Gets mesh from mj at index i."""
  last = (i + 1) >= mj.nmesh
  face_start = mj.mesh_faceadr[i]
  face_end = mj.mesh_faceadr[i + 1] if not last else mj.mesh_face.shape[0]
  face = mj.mesh_face[face_start:face_end]

  vert_start = mj.mesh_vertadr[i]
  vert_end = mj.mesh_vertadr[i + 1] if not last else mj.mesh_vert.shape[0]
  vert = mj.mesh_vert[vert_start:vert_end]

  return vert, face


def _get_name(mj: mujoco.MjModel, i: int) -> str:
  names = mj.names[i:].decode('utf-8')
  return names[: names.find('\x00')]


def _check_custom(mj: mujoco.MjModel, custom: Dict[str, np.ndarray]) -> None:
  """Validates fields in custom."""
  if not (
      0 <= custom['spring_mass_scale'] <= 1
      and 0 <= custom['spring_inertia_scale'] <= 1
  ):
    raise ValueError('Spring inertia and mass scale must be in [0, 1].')
  if 'init_qpos' in custom and custom['init_qpos'].shape[0] != mj.nq:
    size = custom['init_qpos'].shape[0]
    raise ValueError(
        f'init_qpos had length {size} but expected length {mj.nq}.'
    )


def _get_custom(mj: mujoco.MjModel) -> Dict[str, np.ndarray]:
  """Gets custom mjcf parameters for brax, with defaults."""
  default = {
      'ang_damping': (0.0, None),
      'vel_damping': (0.0, None),
      'baumgarte_erp': (0.1, None),
      'spring_mass_scale': (0.0, None),
      'spring_inertia_scale': (0.0, None),
      'joint_scale_pos': (0.5, None),
      'joint_scale_ang': (0.2, None),
      'collide_scale': (1.0, None),
      'matrix_inv_iterations': (10, None),
      'solver_maxls': (20, None),
      'elasticity': (0.0, 'geom'),
      'convex': (True, 'geom'),
      'constraint_stiffness': (2000.0, 'body'),
      'constraint_limit_stiffness': (1000.0, 'body'),
      'constraint_ang_damping': (0.0, 'body'),
      'constraint_vel_damping': (0.0, 'body'),
  }

  # add user provided overrides to the defaults
  for i, ni in enumerate(mj.name_numericadr):
    nsize = mj.numeric_size[i]
    name = _get_name(mj, ni)
    val = mj.numeric_data[mj.numeric_adr[i] : mj.numeric_adr[i] + nsize]
    typ = default[name][1] if name in default else None
    default[name] = (val, typ)

  # gather custom overrides with correct sizes
  custom = {}
  for name, (val, typ) in default.items():
    val = np.array([val])
    size = {
        'body': mj.nbody - 1,  # ignore the world body
        'geom': mj.ngeom,
    }.get(typ, val.shape[-1])
    if val.shape[-1] != size and val.shape[-1] > 1:
      # the provided shape does not match against our default size
      raise ValueError(
          f'"{name}" custom arg needed {size} values for the "{typ}" type, '
          f'but got {val.shape[-1]} values.'
      )
    elif val.shape[-1] != size and val.shape[-1] == 1:
      val = np.repeat(val, size)
    val = val.squeeze() if not typ else val.reshape(size)
    if typ == 'body':
      # pad one value for the world body, which gets dropped at Link creation
      val = np.concatenate([[val[0]], val])
    custom[name] = val

  # get tuple custom overrides
  for i, ni in enumerate(mj.name_tupleadr):
    start, end = mj.tuple_adr[i], mj.tuple_adr[i] + mj.tuple_size[i]
    objtype = mj.tuple_objtype[start:end]
    name = _get_name(mj, ni)
    if not all(objtype[0] == objtype):
      raise NotImplementedError(
          f'All tuple elements "{name}" should have the same object type.'
      )
    if objtype[0] not in [1, 5]:
      raise NotImplementedError(
          f'Custom tuple "{name}" with objtype=={objtype[0]} is not supported.'
      )
    typ = {1: 'body', 5: 'geom'}[objtype[0]]
    if name in default and default[name][1] != typ:
      raise ValueError(
          f'Custom tuple "{name}" is expected to be associated with'
          f' the {default[name][1]} objtype.'
      )

    size = {1: mj.nbody, 5: mj.ngeom}[objtype[0]]
    default_val, _ = default.get(name, (0.0, None))
    arr = np.repeat(default_val, size)
    objid = mj.tuple_objid[start:end]
    objprm = mj.tuple_objprm[start:end]
    arr[objid] = objprm
    custom[name] = arr

  _check_custom(mj, custom)
  return custom


def load_model(mj: mujoco.MjModel) -> System:
  """Creates a brax system from a MuJoCo model."""
  # do some validation up front
  if any(i not in [0, 1] for i in mj.actuator_biastype):
    raise NotImplementedError('Only actuator_biastype in [0, 1] are supported.')
  if any(i != 0 for i in mj.actuator_gaintype):
    raise NotImplementedError('Only actuator_gaintype in [0] is supported.')
  if mj.opt.integrator != 0:
    raise NotImplementedError('Only euler integration is supported.')
  if mj.opt.cone != 0:
    raise NotImplementedError('Only pyramidal cone friction is supported.')
  if not (mj.actuator_trntype == 0).all():
    raise NotImplementedError(
        'Only joint transmission types are supported for actuators.'
    )
  if (mj.geom_solmix[0] != mj.geom_solmix).any():
    raise NotImplementedError('geom_solmix parameter not supported.')
  if (mj.geom_priority[0] != mj.geom_priority).any():
    raise NotImplementedError('geom_priority parameter not supported.')
  if mj.opt.collision == 1:
    raise NotImplementedError('Predefined collisions not supported.')
  q_width = {0: 7, 1: 4, 2: 1, 3: 1}
  non_free = np.concatenate([[j != 0] * q_width[j] for j in mj.jnt_type])
  if mj.qpos0[non_free].any():
    raise NotImplementedError(
        'The `ref` attribute on joint types is not supported.')
  if (mj.geom_fluid != 0).any():
    raise NotImplementedError('Ellipsoid fluid model not implemented.')
  if mj.opt.wind.any():
    raise NotImplementedError('option.wind is not implemented.')
  custom = _get_custom(mj)

  # create links
  joint_positions = [np.array([0.0, 0.0, 0.0])]
  for _, group in itertools.groupby(
      zip(mj.jnt_bodyid, mj.jnt_pos), key=lambda x: x[0]
  ):
    position = np.array([p for _, p in group])
    if not (position == position[0]).all():
      raise RuntimeError('invalid joint stack: only one joint position allowed')
    joint_positions.append(position[0])
  joint_position = np.array(joint_positions)
  identity = np.tile(np.array([1.0, 0.0, 0.0, 0.0]), (mj.nbody, 1))
  link = Link(  # pytype: disable=wrong-arg-types  # jax-ndarray
      transform=Transform(pos=mj.body_pos, rot=mj.body_quat),  # pytype: disable=wrong-arg-types  # jax-ndarray
      inertia=Inertia(  # pytype: disable=wrong-arg-types  # jax-ndarray
          transform=Transform(pos=mj.body_ipos, rot=mj.body_iquat),  # pytype: disable=wrong-arg-types  # jax-ndarray
          i=np.array([np.diag(i) for i in mj.body_inertia]),
          mass=mj.body_mass,
      ),
      invweight=mj.body_invweight0[:, 0],
      joint=Transform(pos=joint_position, rot=identity),  # pytype: disable=wrong-arg-types  # jax-ndarray
      constraint_stiffness=custom['constraint_stiffness'],
      constraint_vel_damping=custom['constraint_vel_damping'],
      constraint_limit_stiffness=custom['constraint_limit_stiffness'],
      constraint_ang_damping=custom['constraint_ang_damping'],
  )
  # skip link 0 which is the world body in mujoco
  link = jax.tree_map(lambda x: x[1:], link)

  # create dofs
  mj.jnt_range[~(mj.jnt_limited == 1), :] = np.array([-np.inf, np.inf])
  motions, limits, stiffnesses = [], [], []
  for typ, axis, limit, stiffness in zip(
      mj.jnt_type, mj.jnt_axis, mj.jnt_range, mj.jnt_stiffness
  ):
    if typ == 0:
      motion = Motion(ang=np.eye(6, 3, -3), vel=np.eye(6, 3))
      limit = np.array([-np.inf] * 6), np.array([np.inf] * 6)
      if stiffness > 0:
        raise RuntimeError('brax does not support stiffness for free joints')
      stiffness = np.zeros(6)
    elif typ == 1:
      motion = Motion(ang=np.eye(3), vel=np.zeros((3, 3)))
      if np.any(~np.isinf(limit)):
        raise RuntimeError('brax does not support joint ranges for ball joints')
      limit = np.array([-np.inf] * 3), np.array([np.inf] * 3)
      stiffness = np.zeros(3)
    elif typ == 2:
      motion = Motion(ang=np.zeros((1, 3)), vel=axis.reshape((1, 3)))
      limit = limit[0:1], limit[1:2]
      stiffness = np.array([stiffness])
    elif typ == 3:
      motion = Motion(ang=axis.reshape((1, 3)), vel=np.zeros((1, 3)))
      limit = limit[0:1], limit[1:2]
      stiffness = np.array([stiffness])
    else:
      raise RuntimeError(f'invalid joint type: {typ}')
    motions.append(motion)
    limits.append(limit)
    stiffnesses.append(stiffness)
  motion = jax.tree_map(lambda *x: np.concatenate(x), *motions)

  limit = None
  if np.any(mj.jnt_limited):
    limit = jax.tree_map(lambda *x: np.concatenate(x), *limits)
  stiffness = np.concatenate(stiffnesses)
  solver_params_jnt = np.concatenate((mj.jnt_solref, mj.jnt_solimp), axis=1)
  solver_params_dof = solver_params_jnt[mj.dof_jntid]

  dof = DoF(  # pytype: disable=wrong-arg-types
      motion=motion,
      armature=mj.dof_armature,
      stiffness=stiffness,
      damping=mj.dof_damping,
      limit=limit,
      invweight=mj.dof_invweight0,
      solver_params=solver_params_dof,
  )

  solver_params_geom = np.concatenate((mj.geom_solref, mj.geom_solimp), axis=1)
  # group geoms so that they can be stacked.  two geoms can be stacked if:
  # - they have the same type
  # - their fields have the same shape (e.g. Mesh verts might vary)
  # - they have the same mask
  key_fn = lambda g, m: (jax.tree_map(np.shape, g), m)
  geom_groups = {}
  for i, typ in enumerate(mj.geom_type):
    rgba = mj.geom_rgba[i]
    if (mj.geom_rgba[i] == [0.5, 0.5, 0.5, 1.0]).all():
      # convert the default mjcf color to brax default color
      rgba = np.array([0.4, 0.33, 0.26, 1.0])
    kwargs = {
        'link_idx': mj.geom_bodyid[i] - 1 if mj.geom_bodyid[i] > 0 else None,
        'transform': Transform(pos=mj.geom_pos[i], rot=mj.geom_quat[i]),
        'friction': mj.geom_friction[i, 0],
        'elasticity': custom['elasticity'][i],
        'solver_params': solver_params_geom[i],
        'rgba': rgba,
    }
    mask = mj.geom_contype[i] | mj.geom_conaffinity[i] << 32
    if typ == 0:  # Plane
      geom = Plane(**kwargs)
      geom_groups.setdefault(key_fn(geom, mask), []).append(geom)
    elif typ == 5:  # Cylinder
      radius, halflength = mj.geom_size[i, 0:2]
      if halflength > 0.001 and mask > 0:
        # TODO: support cylinders with depth.
        raise NotImplementedError(
            'Cylinders of half-length>0.001 are not supported for collision.')
      geom = Cylinder(radius=radius, length=halflength * 2, **kwargs)
      geom_groups.setdefault(key_fn(geom, mask), []).append(geom)
    elif typ == 2:  # Sphere
      geom = Sphere(radius=mj.geom_size[i, 0], **kwargs)
      geom_groups.setdefault(key_fn(geom, mask), []).append(geom)
    elif typ == 3:  # Capsule
      radius, halflength = mj.geom_size[i, 0:2]
      geom = Capsule(radius=radius, length=halflength * 2, **kwargs)
      geom_groups.setdefault(key_fn(geom, mask), []).append(geom)
    elif typ == 6:  # Box
      geom = Box(halfsize=mj.geom_size[i, :], **kwargs)
      geom_groups.setdefault(key_fn(geom, 0), []).append(geom)  # visual only
      if custom['convex'][i]:
        geom = geom_mesh.convex_hull(geom)
      else:
        geom = geom_mesh.box_tri(geom)
      geom_groups.setdefault(key_fn(geom, mask), []).append(geom)
    elif typ == 7:  # Mesh
      vert, face = _get_mesh(mj, mj.geom_dataid[i])
      geom = Mesh(vert=vert, face=face, **kwargs)  # pytype: disable=wrong-arg-types
      if custom['convex'][i]:
        geom_groups.setdefault(key_fn(geom, 0), []).append(geom)  # visual only
        geom = geom_mesh.convex_hull(geom)
      geom_groups.setdefault(key_fn(geom, mask), []).append(geom)
    else:
      warnings.warn(f'unrecognized collider, geom_type: {typ}')
      continue

  geoms = [
      jax.tree_map(lambda *x: np.stack(x), *g) for g in geom_groups.values()
  ]
  geom_masks = [m for _, m in geom_groups.keys()]

  # create actuators
  ctrl_range = mj.actuator_ctrlrange
  ctrl_range[~(mj.actuator_ctrllimited == 1), :] = np.array([-np.inf, np.inf])
  force_range = mj.actuator_forcerange
  force_range[~(mj.actuator_forcelimited == 1), :] = np.array([-np.inf, np.inf])
  q_id = np.array([mj.jnt_qposadr[i] for i in mj.actuator_trnid[:, 0]])
  qd_id = np.array([mj.jnt_dofadr[i] for i in mj.actuator_trnid[:, 0]])
  bias_q = mj.actuator_biasprm[:, 1] * (mj.actuator_biastype != 0)
  bias_qd = mj.actuator_biasprm[:, 2] * (mj.actuator_biastype != 0)

  # TODO: might be nice to add actuator names for debugging
  actuator = Actuator(  # pytype: disable=wrong-arg-types
      q_id=q_id,
      qd_id=qd_id,
      gain=mj.actuator_gainprm[:, 0],
      gear=mj.actuator_gear[:, 0],
      ctrl_range=ctrl_range,
      force_range=force_range,
      bias_q=bias_q,
      bias_qd=bias_qd,
  )

  # create non-pytree params.  these do not live on device directly, and they
  # cannot be differentiated, but they do change the emitted control flow
  link_names = [_get_name(mj, i) for i in mj.name_bodyadr[1:]]
  # convert stacked joints to 1, 2, or 3
  link_types = ''
  for _, group in itertools.groupby(
      zip(mj.jnt_bodyid, mj.jnt_type), key=lambda x: x[0]
  ):
    typs = [t for _, t in group]
    if len(typs) == 1 and typs[0] == 0:  # free
      typ = 'f'
    elif 0 in typs:
      raise RuntimeError('invalid joint stack: cannot stack free joints')
    elif 1 in typs:
      raise NotImplementedError('ball joints not supported')
    else:
      typ = str(len(typs))
    link_types += typ
  link_parents = tuple(mj.body_parentid - 1)[1:]

  # mujoco stores free q in world frame, so clear link transform for free links
  if 'f' in link_types:
    free_idx = np.array([i for i, typ in enumerate(link_types) if typ == 'f'])
    link.transform.pos[free_idx] = np.zeros(3)
    link.transform.rot[free_idx] = np.array([1.0, 0.0, 0.0, 0.0])

  sys = System(  # pytype: disable=wrong-arg-types  # jax-ndarray
      dt=mj.opt.timestep,
      gravity=mj.opt.gravity,
      viscosity=mj.opt.viscosity,
      density=mj.opt.density,
      link=link,
      dof=dof,
      geoms=geoms,
      actuator=actuator,
      init_q=custom['init_qpos'] if 'init_qpos' in custom else mj.qpos0,
      vel_damping=custom['vel_damping'],
      ang_damping=custom['ang_damping'],
      baumgarte_erp=custom['baumgarte_erp'],
      spring_mass_scale=custom['spring_mass_scale'],
      spring_inertia_scale=custom['spring_inertia_scale'],
      joint_scale_ang=custom['joint_scale_ang'],
      joint_scale_pos=custom['joint_scale_pos'],
      collide_scale=custom['collide_scale'],
      enable_fluid=(mj.opt.viscosity > 0) | (mj.opt.density > 0),
      geom_masks=geom_masks,
      link_names=link_names,
      link_types=link_types,
      link_parents=link_parents,
      matrix_inv_iterations=int(custom['matrix_inv_iterations']),
      solver_iterations=mj.opt.iterations,
      solver_maxls=int(custom['solver_maxls']),
  )

  sys = jax.tree_map(jp.array, sys)

  return sys


def fuse_bodies(xml: str):
  """Fuses together parent child bodies that have no joint."""
  xml = ElementTree.fromstring(xml)
  _fuse_bodies(xml)
  return ElementTree.tostring(xml, encoding='unicode')


def loads(xml: str, asset_path: Union[str, epath.Path, None] = None) -> System:
  """Loads a brax system from a MuJoCo mjcf xml string."""
  elem = ElementTree.fromstring(xml)
  _fuse_bodies(elem)
  assets = {}
  if asset_path is not None:
    meshdir = _get_meshdir(elem)
    asset_path = epath.Path(asset_path)
    assets = _find_assets(elem, asset_path, meshdir)
  xml = ElementTree.tostring(elem, encoding='unicode')
  mj = mujoco.MjModel.from_xml_string(xml, assets=assets)

  return load_model(mj)


def load(path: Union[str, epath.Path]):
  """Loads a brax system from a MuJoCo mjcf file path."""
  elem = ElementTree.fromstring(epath.Path(path).read_text())
  _fuse_bodies(elem)
  meshdir = _get_meshdir(elem)
  assets = _find_assets(elem, epath.Path(path), meshdir)
  xml = ElementTree.tostring(elem, encoding='unicode')
  mj = mujoco.MjModel.from_xml_string(xml, assets=assets)

  return load_model(mj)
